# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities and helper functions."""

import contextlib
import functools
import inspect
import pyclbr
import shutil
import sys
import tempfile

from oslo_concurrency import lockutils
from oslo_concurrency import processutils
from oslo_log import log as logging
from oslo_utils import importutils

import masakarimonitors.conf
from masakarimonitors.i18n import _
from masakarimonitors import privsep


CONF = masakarimonitors.conf.CONF

LOG = logging.getLogger(__name__)


def monkey_patch():
    """monkey_patch function.

    If the CONF.monkey_patch set as True,
    this function patches a decorator
    for all functions in specified modules.
    You can set decorators for each modules
    using CONF.monkey_patch_modules.
    The format is "Module path:Decorator function".
    name - name of the function
    function - object of the function
    """
    # If CONF.monkey_patch is not True, this function do nothing.
    if not CONF.monkey_patch:
        return

    def is_method(obj):
        return inspect.ismethod(obj) or inspect.isfunction(obj)
    # Get list of modules and decorators
    for module_and_decorator in CONF.monkey_patch_modules:
        md_value = module_and_decorator.split(':')
        if len(md_value) != 2:
            msg = _("'monkey_patch_modules' config option is not configured "
                    "correctly")
            raise Exception(msg)
        module, decorator_name = md_value
        # import decorator function
        decorator = importutils.import_class(decorator_name)
        __import__(module)
        # Retrieve module information using pyclbr
        module_data = pyclbr.readmodule_ex(module)
        for key, value in module_data.items():
            # set the decorator for the class methods
            if isinstance(value, pyclbr.Class):
                clz = importutils.import_class("%s.%s" % (module, key))
                for method, func in inspect.getmembers(clz, is_method):
                    setattr(clz, method,
                            decorator("%s.%s.%s" % (module, key,
                                                    method), func))
            # set the decorator for the function
            if isinstance(value, pyclbr.Function):
                func = importutils.import_class("%s.%s" % (module, key))
                setattr(sys.modules[module], key,
                        decorator("%s.%s" % (module, key), func))


@contextlib.contextmanager
def tempdir(**kwargs):
    argdict = kwargs.copy()
    if 'dir' not in argdict:
        argdict['dir'] = CONF.tempdir
    tmpdir = tempfile.mkdtemp(**argdict)
    try:
        yield tmpdir
    finally:
        try:
            shutil.rmtree(tmpdir)
        except OSError as e:
            LOG.error('Could not remove tmpdir: %s', e)


@privsep.monitors_priv.entrypoint
def privsep_execute(*cmd, **kwargs):
    return processutils.execute(*cmd, **kwargs)


def execute(*cmd, **kwargs):
    """Convenience wrapper around oslo's execute() method."""
    if 'run_as_root' in kwargs and kwargs.get('run_as_root'):
        return privsep_execute(*cmd, **kwargs)
    else:
        return processutils.execute(*cmd, **kwargs)


def synchronized(name, semaphores=None, blocking=False):
    def wrap(f):
        @functools.wraps(f)
        def inner(*args, **kwargs):
            lock_str = 'masakarimonitors-%s' % name
            int_lock = lockutils.internal_lock(lock_str,
                                               semaphores=semaphores)
            msg = "Lock blocking: %s on resource %s " % (lock_str, f.__name__)
            """Acquiring lock: %(lock_str)s on resource """
            if not int_lock.acquire(blocking=blocking):
                raise Exception(msg)
            try:
                return f(*args, **kwargs)
            finally:
                """Releasing lock: %(lock_str)s on resource """
                int_lock.release()
        return inner
    return wrap



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from xml.etree import ElementTree

from oslo_log import log as oslo_logging

LOG = oslo_logging.getLogger(__name__)


class ParseCibXml(object):
    """ParseCibXml class

    This class parses the cib xml.
    """

    def __init__(self):
        self.cib_tag = None

    def set_cib_xml(self, cib_xml):
        """Set xml.etree.ElementTree.Element object.

        This method recieves string of cib xml, and convert it
        to xml.etree.ElementTree.Element object.

        :params cib_xml: String of cib xml
        """
        # Convert xml.etree.ElementTree.Element object.
        self.cib_tag = ElementTree.fromstring(cib_xml)

    def have_quorum(self):
        """Returns if cluster has quorum or not.

        :returns: 0 on no-quorum, 1 if cluster has quorum.
        """
        return int(self.cib_tag.get('have-quorum'))

    def _get_status_tag(self):
        # status tag exists in the cib tag.
        child_list = list(self.cib_tag)
        for child in child_list:
            if child.tag == 'status':
                return child
        return None

    def _get_node_states(self, status_tag):
        node_state_tag_list = []

        # node_state tag exists in the status tag.
        child_list = list(status_tag)
        for child in child_list:
            if child.tag == 'node_state':
                node_state_tag_list.append(child)

        return node_state_tag_list

    def get_node_state_tag_list(self):
        """Get node_state tag list.

        This method gets node_state tag list from cib xml.

        :returns: node_state tag list
        """
        # Get status tag.
        status_tag = self._get_status_tag()
        if status_tag is None:
            LOG.error("Cib xml doesn't have status tag.")
            return []

        # Get node_state tag list.
        node_state_tag_list = self._get_node_states(status_tag)
        if len(node_state_tag_list) == 0:
            LOG.error("Cib xml doesn't have node_state tag.")

        return node_state_tag_list

    def _parse_instance_attributes_tag(self,
                                       instance_attributes_tag, hostname):
        # Parse nvpair tag under the instance_attributes tag.
        is_target_ipmi = False
        ipmi_values = {}

        nvpair_tag_list = list(instance_attributes_tag)
        for nvpair_tag in nvpair_tag_list:
            if nvpair_tag.get('name') == 'hostname' and \
                nvpair_tag.get('value') == hostname:
                is_target_ipmi = True
            elif nvpair_tag.get('name') == 'ipaddr':
                ipmi_values['ipaddr'] = nvpair_tag.get('value')
            elif nvpair_tag.get('name') == 'userid':
                ipmi_values['userid'] = nvpair_tag.get('value')
            elif nvpair_tag.get('name') == 'passwd':
                ipmi_values['passwd'] = nvpair_tag.get('value')
            elif nvpair_tag.get('name') == 'interface':
                ipmi_values['interface'] = nvpair_tag.get('value')

        if is_target_ipmi is True:
            return ipmi_values
        else:
            return None

    def _parse_primitive_tag(self, primitive_tag, hostname):
        if primitive_tag.get('type') != 'external/ipmi':
            return None

        # Parse instance_attributes tag under the primitive tag.
        child_list = list(primitive_tag)
        for child in child_list:
            if child.tag == 'instance_attributes':
                ipmi_values = self._parse_instance_attributes_tag(
                    child, hostname)
                if ipmi_values is not None:
                    return ipmi_values
        return None

    def _parse_group_tag(self, group_tag, hostname):
        # Parse primitive tag under the group tag.
        child_list = list(group_tag)
        for child in child_list:
            if child.tag == 'primitive':
                ipmi_values = self._parse_primitive_tag(child, hostname)
                if ipmi_values is not None:
                    return ipmi_values
        return None

    def get_stonith_ipmi_params(self, hostname):
        """Get stonith ipmi params from cib xml.

        This method gets params of ipmi resource agent(RA) which is set on
        resources tag.
        The resources tag exists under the configuration tag.
        And it is assumed that ipmi RA belongs to some resource group.

        :params hostname: hostname

        :returns: Dictionary of ipmi RA's params.
                  They are ipaddr, userid, passwd and interface.
        """
        # Get configuration tag from cib tag.
        configuration_tag = None
        child_list = list(self.cib_tag)
        for child in child_list:
            if child.tag == 'configuration':
                configuration_tag = child
                break
        if configuration_tag is None:
            LOG.error("Cib xml doesn't have configuration tag.")
            return None

        # Get resources tag from configuration tag.
        resources_tag = None
        child_list = list(configuration_tag)
        for child in child_list:
            if child.tag == 'resources':
                resources_tag = child
                break
        if resources_tag is None:
            LOG.error("Cib xml doesn't have resources tag.")
            return None

        # They are set at nvpair tag which exists under the
        # instance_attributes of primitive of group tag.
        ipmi_values = None
        child_list = list(resources_tag)
        for child in child_list:
            if child.tag == 'group':
                ipmi_values = self._parse_group_tag(child, hostname)
                if ipmi_values is not None:
                    break

        return ipmi_values



# Copyright (c) 2018 WindRiver Systems
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import eventlet
import libvirt
import socket
import sys
import threading
import time

from oslo_config import cfg
from oslo_log import log as oslo_logging
from oslo_utils import excutils
from oslo_utils import timeutils

from masakarimonitors.introspectiveinstancemonitor import qemu_utils
from masakarimonitors.introspectiveinstancemonitor import scheduler
from masakarimonitors import manager
from masakarimonitors.objects import event_constants as ec

LOG = oslo_logging.getLogger(__name__)
CONF = cfg.CONF


class IntrospectiveInstanceMonitorManager(manager.Manager):

    def __init__(self, *args, **kwargs):
        self.init_tgm()
        super(IntrospectiveInstanceMonitorManager, self).__init__(
            service_name="introspectiveinstancemonitor", *args, **kwargs)
        # This keeps track of what thread is running the event loop,
        # (if it is run in a background thread)
        self.event_loop_thread = None

    def _reset_journal(self, eventID, eventType, detail, uuID):
        """To reset the monitoring to discovery stage

        :param eventID: Event ID
        :param eventType: Event type
        :param detail: Event code
        :param uuID: Domain uuid
        """

        noticeType = ec.EventConstants.TYPE_VM
        hostname = socket.gethostname()
        currentTime = timeutils.utcnow()

        def _reset_handler(event_id, event_type, detail, domain_uuid, msg):
            """Reset monitoring

            To reset monitoring to discovery stage for the following event:
            libvirt.VIR_DOMAIN_EVENT_STARTED
            """

            if (event_id == libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE) and \
               (event_type == libvirt.VIR_DOMAIN_EVENT_STARTED):
                LOG.debug(msg)
                qemu_utils.resetJournal(domain_uuid)

        # All Event Output if debug mode is on.
        msg = "libvirt Event Received.type = %s \
            hostname = %s uuid = %s time = %s eventID = %d eventType = %d \
            detail = %d" % (
            noticeType,
            hostname, uuID, currentTime, eventID,
            eventType, detail)

        LOG.debug("%s", msg)
        try:
            thread = threading.Thread(
                _reset_handler(eventID, eventType, detail, uuID, msg)
            )
            thread.start()
        except KeyError as e:
            LOG.error("virEventFilter KeyError: {0}".format(e))
        except IndexError as e:
            LOG.error("virEventFilter IndexError: {0}".format(e))
        except TypeError as e:
            LOG.error("virEventFilter TypeError: {0}".format(e))
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error("Unexpected error: %s" % sys.exc_info()[0])

    def init_tgm(self):
        """Manages the masakari-introspectiveinstancemonitor."""
        self.TG = scheduler.ThreadGroupManager()

    def _vir_event_loop_native_run(self):
        # Directly run the event loop in the current thread
        while True:
            libvirt.virEventRunDefaultImpl()

    def _vir_event_loop_native_start(self):
        libvirt.virEventRegisterDefaultImpl()
        self.event_loop_thread = threading.Thread(
            target=self._vir_event_loop_native_run,
            name="lib_virt_eventLoop")
        self.event_loop_thread.daemon = True
        self.event_loop_thread.start()

    def _my_domain_event_callback(self, conn, dom, event, detail, opaque):
        self._reset_journal(libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE,
        event, detail, dom.UUIDString())

    def _my_domain_event_reboot_callback(self, conn, dom, opaque):
        self._reset_journal(libvirt.VIR_DOMAIN_EVENT_ID_REBOOT,
        -1, -1, dom.UUIDString())

    def _err_handler(self, ctxt, err):
        LOG.warning("Error from libvirt : %s", err[2])

    def _virt_event(self, uri):
        # Run a background thread with the event loop
        self._vir_event_loop_native_start()

        event_callback_handlers = {
            libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE:
                self._my_domain_event_callback,
            libvirt.VIR_DOMAIN_EVENT_ID_REBOOT:
                self._my_domain_event_reboot_callback
        }
        # Connect to libvirt - If be disconnected, reprocess.
        self.running = True
        while self.running:
            vc = libvirt.openReadOnly(uri)

            # Event callback settings
            callback_ids = []
            for event, callback in event_callback_handlers.items():
                cid = vc.domainEventRegisterAny(None, event, callback, None)
                callback_ids.append(cid)

            # Connection monitoring.
            vc.setKeepAlive(5, 3)
            while vc.isAlive() == 1 and self.running:
                eventlet.greenthread.sleep(1)

            # If connection between libvirtd was lost,
            # clear callback connection.
            LOG.warning("Libvirt Connection Closed Unexpectedly.")
            for cid in callback_ids:
                try:
                    vc.domainEventDeregisterAny(cid)
                except Exception:
                    LOG.debug("Failed to deregister libvirt event "
                              "callback %s.", cid, exc_info=True)
            vc.close()
            del vc
            time.sleep(3)

    def stop(self):
        self.running = False

    def main(self):
        """Main method.

        Set the URI, error handler, and executes event loop processing.
        """

        uri = CONF.libvirt.connection_uri
        LOG.debug("Using uri:" + uri)

        # set error handler & do event loop
        libvirt.registerErrorHandler(self._err_handler, '_virt_event')
        self._virt_event(uri)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
import testtools
from unittest import mock
from xml.etree import ElementTree

from collections import deque
import eventlet
from oslo_utils import timeutils

import masakarimonitors.conf
from masakarimonitors.ha import masakari
from masakarimonitors.hostmonitor.host_handler import handle_host
from masakarimonitors.hostmonitor.host_handler import hold_host_status
from masakarimonitors.hostmonitor.host_handler import parse_cib_xml
from masakarimonitors.hostmonitor.host_handler import parse_crmmon_xml
from masakarimonitors.objects import event_constants as ec
from masakarimonitors import utils

eventlet.monkey_patch(os=False)

CONF = masakarimonitors.conf.CONF
STATUS_TAG_XML = '  <status>' \
                 '    <node_state crmd="online" uname="node1">' \
                 '      <test foo="foo"/>' \
                 '    </node_state>' \
                 '    <node_state crmd="online" uname="node2">' \
                 '      <test foo="foo"/>' \
                 '    </node_state>' \
                 '    <node_state crmd="online" uname="node3">' \
                 '      <test foo="foo"/>' \
                 '    </node_state>' \
                 '    <node_state crmd="offline" uname="node4">' \
                 '      <test foo="foo"/>' \
                 '    </node_state>' \
                 '    <node_state crmd="other" uname="node5">' \
                 '      <test foo="foo"/>' \
                 '    </node_state>' \
                 '  </status>'
CRMMON_NODES_TAG_XML = """
<nodes>
  <node name="member1" id="1002" online="true" standby="false"
        standby_onfail="false" maintenance="false" pending="false"
        unclean="false" shutdown="false" expected_up="true" is_dc="false"
        resources_running="2" type="member" />
  <node name="member2" id="1001" online="true" standby="false"
        standby_onfail="false" maintenance="false" pending="false"
        unclean="false" shutdown="false" expected_up="true" is_dc="true"
        resources_running="1" type="member" />
  <node name="remote1" id="remotehostname1" online="true" standby="false"
        standby_onfail="false" maintenance="false" pending="false"
        unclean="false" shutdown="false" expected_up="false"
        is_dc="false" resources_running="0" type="remote" />
  <node name="remote2" id="remotehostname2" online="true" standby="false"
        standby_onfail="false" maintenance="false" pending="false"
        unclean="false" shutdown="false" expected_up="false" is_dc="false"
        resources_running="0" type="remote" />
  <node name="remote3" id="remotehostname3" online="true" standby="false"
        standby_onfail="false" maintenance="false" pending="false"
        unclean="false" shutdown="false" expected_up="false" is_dc="false"
        resources_running="0" type="remote" />
  <node name="member3" id="1000" online="true" standby="false"
        standby_onfail="false" maintenance="false" pending="false"
        unclean="false" shutdown="false" expected_up="true" is_dc="false"
        resources_running="4" type="member" />
</nodes>
"""


class TestCibSchemaCompliantTag(testtools.TestCase):

    def setUp(self):
        super(TestCibSchemaCompliantTag, self).setUp()

    def test_init_offline(self):
        tag = handle_host.CibSchemaCompliantTag(
            {'name': 'test1', 'online': 'false'})
        self.assertEqual(tag['uname'], 'test1')
        self.assertEqual(tag['crmd'], 'offline')

    def test_init_online(self):
        tag = handle_host.CibSchemaCompliantTag(
            {'name': 'test1', 'online': 'true'})
        self.assertEqual(tag['uname'], 'test1')
        self.assertEqual(tag['crmd'], 'online')


class TestHandleHost(testtools.TestCase):

    def setUp(self):
        super(TestHandleHost, self).setUp()

    @mock.patch.object(utils, 'execute')
    def test_check_pacemaker_services(self, mock_execute):

        mock_execute.return_value = ('test_stdout', '')

        obj = handle_host.HandleHost()
        ret = obj._check_pacemaker_services('corosync')

        self.assertTrue(ret)
        mock_execute.assert_called_once_with(
            'systemctl', 'status', 'corosync', run_as_root=True)

    @mock.patch.object(utils, 'execute')
    def test_check_pacemaker_services_stderr(self, mock_execute):

        mock_execute.return_value = ('test_stdout', 'test_stderr')

        obj = handle_host.HandleHost()
        ret = obj._check_pacemaker_services('corosync')

        self.assertFalse(ret)
        mock_execute.assert_called_once_with(
            'systemctl', 'status', 'corosync', run_as_root=True)

    @mock.patch.object(utils, 'execute')
    @mock.patch.object(handle_host.HandleHost, '_check_pacemaker_services')
    def test_check_hb_line(
        self, mock_check_pacemaker_services, mock_execute):

        mock_check_pacemaker_services.side_effect = [True, True, False]
        interfaces = "enp0s8"
        ports = "5405"
        CONF.set_override("corosync_multicast_interfaces", interfaces, "host")
        CONF.set_override("corosync_multicast_ports", ports, "host")
        mock_execute.return_value = ('', '')

        obj = handle_host.HandleHost()
        ret = obj._check_hb_line()

        self.assertEqual(0, ret)
        self.assertEqual(3, mock_check_pacemaker_services.call_count)
        mock_check_pacemaker_services.assert_any_call('corosync')
        mock_check_pacemaker_services.assert_any_call('pacemaker')
        mock_check_pacemaker_services.assert_any_call('pacemaker_remote')
        cmd_str = ("timeout %s tcpdump -n -c 1 -p -i %s port %s") \
            % (CONF.host.tcpdump_timeout, interfaces, ports)
        command = cmd_str.split()
        mock_execute.assert_called_once_with(*command, run_as_root=True)

    @mock.patch.object(handle_host.HandleHost, '_check_pacemaker_services')
    def test_check_hb_line_pacemaker_not_running(
        self, mock_check_pacemaker_services):

        mock_check_pacemaker_services.side_effect = [False, False, False]

        obj = handle_host.HandleHost()
        ret = obj._check_hb_line()

        self.assertEqual(2, ret)
        self.assertEqual(3, mock_check_pacemaker_services.call_count)
        mock_check_pacemaker_services.assert_any_call('corosync')
        mock_check_pacemaker_services.assert_any_call('pacemaker')
        mock_check_pacemaker_services.assert_any_call('pacemaker_remote')

    @mock.patch.object(handle_host.HandleHost, '_check_pacemaker_services')
    def test_check_hb_line_work_on_pacemaker_remote(
        self, mock_check_pacemaker_services):

        mock_check_pacemaker_services.side_effect = [False, False, True]

        obj = handle_host.HandleHost()
        ret = obj._check_hb_line()

        self.assertEqual(0, ret)
        self.assertEqual(3, mock_check_pacemaker_services.call_count)
        mock_check_pacemaker_services.assert_any_call('corosync')
        mock_check_pacemaker_services.assert_any_call('pacemaker')
        mock_check_pacemaker_services.assert_any_call('pacemaker_remote')

    @mock.patch.object(handle_host.HandleHost, '_check_pacemaker_services')
    def test_check_hb_line_conf_interfaces_is_none(
        self, mock_check_pacemaker_services):

        mock_check_pacemaker_services.side_effect = [True, True, False]
        interfaces = None
        ports = "5405"
        CONF.set_override("corosync_multicast_interfaces", interfaces, "host")
        CONF.set_override("corosync_multicast_ports", ports, "host")

        obj = handle_host.HandleHost()
        ret = obj._check_hb_line()

        self.assertEqual(2, ret)
        self.assertEqual(3, mock_check_pacemaker_services.call_count)
        mock_check_pacemaker_services.assert_any_call('corosync')
        mock_check_pacemaker_services.assert_any_call('pacemaker')
        mock_check_pacemaker_services.assert_any_call('pacemaker_remote')

    @mock.patch.object(handle_host.HandleHost, '_check_pacemaker_services')
    def test_check_hb_line_conf_ports_is_none(
        self, mock_check_pacemaker_services):

        mock_check_pacemaker_services.side_effect = [True, True, False]
        interfaces = "enp0s8"
        ports = None
        CONF.set_override("corosync_multicast_interfaces", interfaces, "host")
        CONF.set_override("corosync_multicast_ports", ports, "host")

        obj = handle_host.HandleHost()
        ret = obj._check_hb_line()

        self.assertEqual(2, ret)
        self.assertEqual(3, mock_check_pacemaker_services.call_count)
        mock_check_pacemaker_services.assert_any_call('corosync')
        mock_check_pacemaker_services.assert_any_call('pacemaker')
        mock_check_pacemaker_services.assert_any_call('pacemaker_remote')

    @mock.patch.object(handle_host.HandleHost, '_check_pacemaker_services')
    def test_check_hb_line_incorrect_interfaces(
        self, mock_check_pacemaker_services):

        mock_check_pacemaker_services.side_effect = [True, True, False]
        interfaces = "enp0s3,enp0s8"
        ports = "5405"
        CONF.set_override("corosync_multicast_interfaces", interfaces, "host")
        CONF.set_override("corosync_multicast_ports", ports, "host")

        obj = handle_host.HandleHost()
        ret = obj._check_hb_line()

        self.assertEqual(2, ret)
        self.assertEqual(3, mock_check_pacemaker_services.call_count)
        mock_check_pacemaker_services.assert_any_call('corosync')
        mock_check_pacemaker_services.assert_any_call('pacemaker')
        mock_check_pacemaker_services.assert_any_call('pacemaker_remote')

    @mock.patch.object(handle_host.HandleHost, '_check_pacemaker_services')
    def test_check_hb_line_incorrect_ports(
        self, mock_check_pacemaker_services):

        mock_check_pacemaker_services.side_effect = [True, True, False]
        interfaces = "enp0s8"
        ports = "5405,5406"
        CONF.set_override("corosync_multicast_interfaces", interfaces, "host")
        CONF.set_override("corosync_multicast_ports", ports, "host")

        obj = handle_host.HandleHost()
        ret = obj._check_hb_line()

        self.assertEqual(2, ret)
        self.assertEqual(3, mock_check_pacemaker_services.call_count)
        mock_check_pacemaker_services.assert_any_call('corosync')
        mock_check_pacemaker_services.assert_any_call('pacemaker')
        mock_check_pacemaker_services.assert_any_call('pacemaker_remote')

    @mock.patch.object(utils, 'execute')
    @mock.patch.object(handle_host.HandleHost, '_check_pacemaker_services')
    def test_check_hb_line_tcpdump_fail(
        self, mock_check_pacemaker_services, mock_execute):

        mock_check_pacemaker_services.side_effect = [True, True, False]
        interfaces = "enp0s8"
        ports = "5405"
        CONF.set_override("corosync_multicast_interfaces", interfaces, "host")
        CONF.set_override("corosync_multicast_ports", ports, "host")

        mock_execute.side_effect = Exception("Test exception.")

        obj = handle_host.HandleHost()
        ret = obj._check_hb_line()

        self.assertEqual(1, ret)
        self.assertEqual(3, mock_check_pacemaker_services.call_count)
        mock_check_pacemaker_services.assert_any_call('corosync')
        mock_check_pacemaker_services.assert_any_call('pacemaker')
        mock_check_pacemaker_services.assert_any_call('pacemaker_remote')
        cmd_str = ("timeout %s tcpdump -n -c 1 -p -i %s port %s") \
            % (CONF.host.tcpdump_timeout, interfaces, ports)
        command = cmd_str.split()
        mock_execute.assert_called_once_with(*command, run_as_root=True)

    @mock.patch.object(utils, 'execute')
    def test_check_host_status_by_crmadmin_idle(self, mock_execute):
        my_hostname = socket.gethostname()
        crmadmin_stdout = "Status of crmd@%s: S_IDLE (ok)" % my_hostname
        mock_execute.return_value = (crmadmin_stdout, '')

        obj = handle_host.HandleHost()
        ret = obj._check_host_status_by_crmadmin()

        self.assertEqual(0, ret)
        mock_execute.assert_called_once_with(
            'crmadmin', '-S', my_hostname, run_as_root=True)

    @mock.patch.object(utils, 'execute')
    def test_check_host_status_by_crmadmin_not_dc(self, mock_execute):
        my_hostname = socket.gethostname()
        crmadmin_stdout = "Status of crmd@%s: S_NOT_DC (ok)" % my_hostname
        mock_execute.return_value = (crmadmin_stdout, '')

        obj = handle_host.HandleHost()
        ret = obj._check_host_status_by_crmadmin()

        self.assertEqual(0, ret)
        mock_execute.assert_called_once_with(
            'crmadmin', '-S', my_hostname, run_as_root=True)

    @mock.patch.object(utils, 'execute')
    def test_check_host_status_by_crmadmin_output_stderr(self, mock_execute):
        my_hostname = socket.gethostname()
        mock_execute.return_value = ('', 'crmadmin: command not found')

        obj = handle_host.HandleHost()
        ret = obj._check_host_status_by_crmadmin()

        self.assertEqual(1, ret)
        mock_execute.assert_called_once_with(
            'crmadmin', '-S', my_hostname, run_as_root=True)

    @mock.patch.object(utils, 'execute')
    def test_check_host_status_by_crmadmin_output_unexpected_message(
        self, mock_execute):
        my_hostname = socket.gethostname()
        mock_execute.return_value = ('Unexpected message.', '')

        obj = handle_host.HandleHost()
        ret = obj._check_host_status_by_crmadmin()

        self.assertEqual(1, ret)
        mock_execute.assert_called_once_with(
            'crmadmin', '-S', my_hostname, run_as_root=True)

    @mock.patch.object(utils, 'execute')
    def test_get_cib_xml(self, mock_execute):
        mock_execute.return_value = ('test_stdout', '')

        obj = handle_host.HandleHost()
        ret = obj._get_cib_xml()

        self.assertEqual('test_stdout', ret)
        mock_execute.assert_called_once_with(
            'cibadmin', '--query', run_as_root=True)

    @mock.patch.object(utils, 'execute')
    def test_get_cib_xml_output_stderr(self, mock_execute):
        mock_execute.return_value = ('test_stdout', 'test_stderr')

        obj = handle_host.HandleHost()
        ret = obj._get_cib_xml()

        self.assertIsNone(ret)
        mock_execute.assert_called_once_with(
            'cibadmin', '--query', run_as_root=True)

    @mock.patch.object(utils, 'execute')
    def test_get_crmmon_xml(self, mock_execute):
        mock_execute.return_value = ('test_stdout', '')

        obj = handle_host.HandleHost()
        ret = obj._get_crmmon_xml()

        self.assertEqual('test_stdout', ret)
        mock_execute.assert_called_once_with(
            'crm_mon', '-X', run_as_root=True)

    @mock.patch.object(utils, 'execute')
    def test_get_crmmon_xml_stderr(self, mock_execute):
        mock_execute.return_value = ('test_stdout', 'test_stderr')

        obj = handle_host.HandleHost()
        ret = obj._get_crmmon_xml()

        self.assertIsNone(ret)
        mock_execute.assert_called_once_with(
            'crm_mon', '-X', run_as_root=True)

    @mock.patch.object(utils, 'execute')
    @mock.patch.object(parse_cib_xml.ParseCibXml, 'get_stonith_ipmi_params')
    def test_is_poweroff(self, mock_get_stonith_ipmi_params, mock_execute):
        ipmi_values = {
            'userid': 'admin',
            'passwd': 'password',
            'interface': 'lanplus',
            'ipaddr': '0.0.0.0'
        }
        mock_get_stonith_ipmi_params.return_value = ipmi_values
        mock_execute.return_value = ('Chassis Power is off', '')

        obj = handle_host.HandleHost()
        ret = obj._is_poweroff('test_hostname')

        self.assertTrue(ret)
        cmd_str = ("timeout %s ipmitool -U %s -P %s -I %s -H %s "
                   "power status") \
            % (str(CONF.host.ipmi_timeout), ipmi_values['userid'],
               ipmi_values['passwd'], ipmi_values['interface'],
               ipmi_values['ipaddr'])
        command = cmd_str.split()
        mock_execute.assert_called_once_with(*command, run_as_root=False)

    @mock.patch.object(utils, 'execute')
    @mock.patch.object(parse_cib_xml.ParseCibXml, 'get_stonith_ipmi_params')
    def test_is_poweroff_ipmi_values_is_none(
        self, mock_get_stonith_ipmi_params, mock_execute):
        mock_get_stonith_ipmi_params.return_value = None
        mock_execute.return_value = ('Chassis Power is off', '')

        obj = handle_host.HandleHost()
        ret = obj._is_poweroff('test_hostname')

        self.assertFalse(ret)
        mock_execute.assert_not_called()

    @mock.patch.object(eventlet.greenthread, 'sleep')
    @mock.patch.object(utils, 'execute')
    @mock.patch.object(parse_cib_xml.ParseCibXml, 'get_stonith_ipmi_params')
    def test_is_poweroff_output_stderr(
        self, mock_get_stonith_ipmi_params, mock_execute, mock_sleep):
        ipmi_values = {
            'userid': 'admin',
            'passwd': 'password',
            'interface': 'lanplus',
            'ipaddr': '0.0.0.0'
        }
        mock_get_stonith_ipmi_params.return_value = ipmi_values
        mock_execute.return_value = ('Chassis Power is off', 'test_stderr')
        mock_sleep.return_value = None

        obj = handle_host.HandleHost()
        ret = obj._is_poweroff('test_hostname')

        self.assertFalse(ret)
        cmd_str = ("timeout %s ipmitool -U %s -P %s -I %s -H %s "
                   "power status") \
            % (str(CONF.host.ipmi_timeout), ipmi_values['userid'],
               ipmi_values['passwd'], ipmi_values['interface'],
               ipmi_values['ipaddr'])
        command = cmd_str.split()
        calls = [mock.call(*command, run_as_root=False),
                 mock.call(*command, run_as_root=False),
                 mock.call(*command, run_as_root=False),
                 mock.call(*command, run_as_root=False)]
        mock_execute.assert_has_calls(calls)

    @mock.patch.object(eventlet.greenthread, 'sleep')
    @mock.patch.object(utils, 'execute')
    @mock.patch.object(parse_cib_xml.ParseCibXml, 'get_stonith_ipmi_params')
    def test_is_poweroff_output_unexpected_message(
        self, mock_get_stonith_ipmi_params, mock_execute, mock_sleep):
        ipmi_values = {
            'userid': 'admin',
            'passwd': 'password',
            'interface': 'lanplus',
            'ipaddr': '0.0.0.0'
        }
        mock_get_stonith_ipmi_params.return_value = ipmi_values
        mock_execute.return_value = ('Unexpected message.', '')
        mock_sleep.return_value = None

        obj = handle_host.HandleHost()
        ret = obj._is_poweroff('test_hostname')

        self.assertFalse(ret)
        cmd_str = ("timeout %s ipmitool -U %s -P %s -I %s -H %s "
                   "power status") \
            % (str(CONF.host.ipmi_timeout), ipmi_values['userid'],
               ipmi_values['passwd'], ipmi_values['interface'],
               ipmi_values['ipaddr'])
        command = cmd_str.split()
        calls = [mock.call(*command, run_as_root=False),
                 mock.call(*command, run_as_root=False),
                 mock.call(*command, run_as_root=False),
                 mock.call(*command, run_as_root=False)]
        mock_execute.assert_has_calls(calls)

    @mock.patch.object(timeutils, 'utcnow')
    def test_make_event_online(self, mock_utcnow):
        current_time = timeutils.utcnow()
        mock_utcnow.return_value = current_time

        obj = handle_host.HandleHost()
        hostname = 'test_hostname'
        current_status = 'online'
        ret = obj._make_event(hostname, current_status)

        event_type = ec.EventConstants.EVENT_STARTED
        cluster_status = current_status.upper()
        host_status = ec.EventConstants.HOST_STATUS_NORMAL
        event = {
            'notification': {
                'type': ec.EventConstants.TYPE_COMPUTE_HOST,
                'hostname': hostname,
                'generated_time': current_time,
                'payload': {
                    'event': event_type,
                    'cluster_status': cluster_status,
                    'host_status': host_status
                }
            }
        }
        self.assertEqual(event, ret)

    @mock.patch.object(timeutils, 'utcnow')
    @mock.patch.object(handle_host.HandleHost, '_is_poweroff')
    def test_make_event_offline_disable_ipmi_check_false_poweroff(
        self, mock_is_poweroff, mock_utcnow):
        mock_is_poweroff.return_value = True
        current_time = timeutils.utcnow()
        mock_utcnow.return_value = current_time
        CONF.host.disable_ipmi_check = False

        obj = handle_host.HandleHost()
        hostname = 'test_hostname'
        current_status = 'offline'
        ret = obj._make_event(hostname, current_status)

        event_type = ec.EventConstants.EVENT_STOPPED
        cluster_status = current_status.upper()
        host_status = ec.EventConstants.HOST_STATUS_NORMAL
        event = {
            'notification': {
                'type': ec.EventConstants.TYPE_COMPUTE_HOST,
                'hostname': hostname,
                'generated_time': current_time,
                'payload': {
                    'event': event_type,
                    'cluster_status': cluster_status,
                    'host_status': host_status
                }
            }
        }
        self.assertEqual(event, ret)
        mock_is_poweroff.assert_called_once_with(hostname)

    @mock.patch.object(timeutils, 'utcnow')
    @mock.patch.object(handle_host.HandleHost, '_is_poweroff')
    def test_make_event_offline_disable_ipmi_check_false_poweron(
        self, mock_is_poweroff, mock_utcnow):
        mock_is_poweroff.return_value = False
        current_time = timeutils.utcnow()
        mock_utcnow.return_value = current_time
        CONF.host.disable_ipmi_check = False

        obj = handle_host.HandleHost()
        hostname = 'test_hostname'
        current_status = 'offline'
        ret = obj._make_event(hostname, current_status)

        event_type = ec.EventConstants.EVENT_STOPPED
        cluster_status = current_status.upper()
        host_status = ec.EventConstants.HOST_STATUS_UNKNOWN
        event = {
            'notification': {
                'type': ec.EventConstants.TYPE_COMPUTE_HOST,
                'hostname': hostname,
                'generated_time': current_time,
                'payload': {
                    'event': event_type,
                    'cluster_status': cluster_status,
                    'host_status': host_status
                }
            }
        }
        self.assertEqual(event, ret)
        mock_is_poweroff.assert_called_once_with(hostname)

    @mock.patch.object(timeutils, 'utcnow')
    def test_make_event_offline_disable_ipmi_check_true(self, mock_utcnow):
        current_time = timeutils.utcnow()
        mock_utcnow.return_value = current_time
        CONF.host.disable_ipmi_check = True

        obj = handle_host.HandleHost()
        hostname = 'test_hostname'
        current_status = 'offline'
        ret = obj._make_event(hostname, current_status)

        event_type = ec.EventConstants.EVENT_STOPPED
        cluster_status = current_status.upper()
        host_status = ec.EventConstants.HOST_STATUS_NORMAL
        event = {
            'notification': {
                'type': ec.EventConstants.TYPE_COMPUTE_HOST,
                'hostname': hostname,
                'generated_time': current_time,
                'payload': {
                    'event': event_type,
                    'cluster_status': cluster_status,
                    'host_status': host_status
                }
            }
        }
        self.assertEqual(event, ret)

    def test_normalize_host_status_pre_pacemaker_pre_2_1_7(self):
        status_xml = ElementTree.fromstring("""
            <status>
                <node_state crmd="online" uname="node1" />
                <node_state crmd="offline" uname="node2" />
                <node_state crmd="other" uname="node3" />
            </status>
        """)

        obj = handle_host.HandleHost()
        response = [obj._normalize_host_status(status.get('crmd'))
                    for status in list(status_xml)]
        self.assertEqual(["online", "offline", "other"], response)

    def test_normalize_host_status_pacemaker_post_2_1_7(self):
        status_tag = ElementTree.fromstring("""
            <status>
                <node_state crmd="1767856436" uname="node1" />
                <node_state crmd="0" uname="node2" />
            </status>
        """)

        obj = handle_host.HandleHost()
        response = [obj._normalize_host_status(status.get('crmd'))
                    for status in list(status_tag)]
        self.assertEqual(["online", "offline"], response)

    @mock.patch.object(masakari.SendNotification, 'send_notification')
    @mock.patch.object(handle_host.HandleHost, '_make_event')
    @mock.patch.object(hold_host_status.HostHoldStatus, 'set_host_status')
    @mock.patch.object(hold_host_status.HostHoldStatus, 'get_host_status')
    @mock.patch.object(socket, 'gethostname')
    def test_check_if_status_changed(
        self, mock_gethostname, mock_get_host_status, mock_set_host_status,
        mock_make_event, mock_send_notification):
        mock_gethostname.return_value = 'node1'
        mock_get_host_status.side_effect = \
            [None, 'online', 'online', 'online']
        mock_set_host_status.return_value = None
        test_event = {'notification': 'test'}
        mock_make_event.return_value = test_event

        status_tag = ElementTree.fromstring(STATUS_TAG_XML)
        node_state_tag_list = list(status_tag)

        obj = handle_host.HandleHost()
        obj._check_if_status_changed(node_state_tag_list)

        node_state_node2 = node_state_tag_list[1]
        node_state_node3 = node_state_tag_list[2]
        node_state_node4 = node_state_tag_list[3]
        node_state_node5 = node_state_tag_list[4]
        node2 = node_state_node2.get('uname')
        node3 = node_state_node3.get('uname')
        node4 = node_state_node4.get('uname')
        node5 = node_state_node5.get('uname')
        calls_get_host_status = [mock.call(node2),
                                 mock.call(node3),
                                 mock.call(node4),
                                 mock.call(node5)]
        calls_set_host_status = [mock.call(node_state_node4),
                                 mock.call(node_state_node5)]
        mock_get_host_status.assert_has_calls(calls_get_host_status)
        mock_set_host_status.assert_has_calls(calls_set_host_status)
        mock_make_event.assert_called_once_with(node4, 'offline')
        mock_send_notification.assert_called_once_with(
            CONF.host.api_retry_max, CONF.host.api_retry_interval, test_event)

    @mock.patch.object(masakari.SendNotification, 'send_notification')
    @mock.patch.object(handle_host.HandleHost, '_make_event')
    @mock.patch.object(hold_host_status.HostHoldStatus, 'set_host_status')
    @mock.patch.object(hold_host_status.HostHoldStatus, 'get_host_status')
    @mock.patch.object(socket, 'gethostname')
    def test_check_if_status_changed_with_3_samples(
        self, mock_gethostname, mock_get_host_status, mock_set_host_status,
        mock_make_event, mock_send_notification):
        mock_gethostname.return_value = 'node1'
        mock_get_host_status.side_effect = \
            [None, 'online', 'online', 'online']
        mock_set_host_status.return_value = None
        test_event = {'notification': 'test'}
        mock_make_event.return_value = test_event

        status_tag = ElementTree.fromstring(STATUS_TAG_XML)
        node_state_tag_list = list(status_tag)
        CONF.host.monitoring_samples = 3

        obj = handle_host.HandleHost()
        obj.monitoring_data = {
            "node1": deque(['online', 'online', 'online'], maxlen=3),
            "node2": deque(['offline', 'online', 'online'], maxlen=3),
            "node3": deque(['offline'], maxlen=3),
            "node4": deque(['online', 'offline', 'offline'], maxlen=3),
            "node5": deque(['online', 'online', 'online'], maxlen=3),
        }
        obj._check_if_status_changed(node_state_tag_list)

        self.assertEqual(deque(['online', 'online', 'online'], maxlen=3),
                         obj.monitoring_data.get('node2'))
        self.assertEqual('online', obj.get_stabilised_host_status('node2'))
        self.assertIn(mock.call(node_state_tag_list[1]),
                      mock_set_host_status.mock_calls)

        self.assertEqual(deque(['offline', 'online'], maxlen=3),
                         obj.monitoring_data.get('node3'))
        self.assertEqual('_being_collected',
                         obj.get_stabilised_host_status('node3'))
        self.assertNotIn(mock.call(node_state_tag_list[2]),
                         mock_set_host_status.mock_calls)

        self.assertEqual(deque(['offline', 'offline', 'offline'], maxlen=3),
                         obj.monitoring_data.get('node4'))
        self.assertEqual('offline', obj.get_stabilised_host_status('node4'))
        self.assertIn(mock.call(node_state_tag_list[3]),
                      mock_set_host_status.mock_calls)

        self.assertEqual(deque(['online', 'online', 'other'], maxlen=3),
                         obj.monitoring_data.get('node5'))
        self.assertEqual('_uncertain', obj.get_stabilised_host_status('node5'))
        self.assertNotIn(mock.call(node_state_tag_list[4]),
                         mock_set_host_status.mock_calls)

        mock_make_event.assert_called_once_with("node4", 'offline')
        mock_send_notification.assert_called_once_with(
            CONF.host.api_retry_max, CONF.host.api_retry_interval, test_event)

    @mock.patch.object(handle_host.HandleHost, '_check_if_status_changed')
    @mock.patch.object(parse_crmmon_xml.ParseCrmMonXml,
                       'get_node_state_tag_list')
    @mock.patch.object(parse_crmmon_xml.ParseCrmMonXml,
                       'has_quorum')
    @mock.patch.object(parse_crmmon_xml.ParseCrmMonXml, 'set_crmmon_xml')
    @mock.patch.object(handle_host.HandleHost, '_get_crmmon_xml')
    def test_check_host_status_by_crm_mon(
        self, mock_get_crmmon_xml, mock_set_crmmon_xml,
        mock_has_quorum, mock_get_node_state_tag_list,
        mock_check_if_status_changed):
        mock_get_crmmon_xml.return_value = CRMMON_NODES_TAG_XML
        mock_set_crmmon_xml.return_value = None
        mock_has_quorum.return_value = True
        status_tag = ElementTree.fromstring(CRMMON_NODES_TAG_XML)
        node_state_tag_list = list(status_tag)
        mock_get_node_state_tag_list.return_value = node_state_tag_list
        mock_check_if_status_changed.return_value = None

        obj = handle_host.HandleHost()
        ret = obj._check_host_status_by_crm_mon()

        self.assertEqual(0, ret)
        mock_set_crmmon_xml.assert_called_once_with(CRMMON_NODES_TAG_XML)
        mock_get_node_state_tag_list.assert_called_once_with()
        mock_check_if_status_changed.assert_called_once_with(
            [
                {'uname': 'remote1', 'crmd': 'online'},
                {'uname': 'remote2', 'crmd': 'online'},
                {'uname': 'remote3', 'crmd': 'online'}])

    @mock.patch.object(handle_host.HandleHost, '_check_if_status_changed')
    @mock.patch.object(parse_crmmon_xml.ParseCrmMonXml,
                       'has_quorum')
    @mock.patch.object(parse_crmmon_xml.ParseCrmMonXml, 'set_crmmon_xml')
    @mock.patch.object(handle_host.HandleHost, '_get_crmmon_xml')
    def test_check_host_status_by_crm_mon_no_quorum(
        self, mock_get_crmmon_xml, mock_set_crmmon_xml,
        mock_has_quorum, mock_check_if_status_changed):
        mock_get_crmmon_xml.return_value = CRMMON_NODES_TAG_XML
        mock_set_crmmon_xml.return_value = None
        mock_has_quorum.return_value = False

        obj = handle_host.HandleHost()
        ret = obj._check_host_status_by_crm_mon()

        self.assertEqual(2, ret)
        mock_set_crmmon_xml.assert_called_once_with(CRMMON_NODES_TAG_XML)
        mock_check_if_status_changed.assert_not_called()

    @mock.patch.object(parse_crmmon_xml.ParseCrmMonXml,
                       'has_quorum')
    @mock.patch.object(parse_crmmon_xml.ParseCrmMonXml,
                       'get_node_state_tag_list')
    @mock.patch.object(parse_crmmon_xml.ParseCrmMonXml, 'set_crmmon_xml')
    @mock.patch.object(handle_host.HandleHost, '_get_crmmon_xml')
    def test_check_host_status_by_crm_mon_not_have_node_state_tag(
        self, mock_get_crmmon_xml, mock_set_crmmon_xml,
        mock_get_node_state_tag_list, mock_has_quorum):
        mock_get_crmmon_xml.return_value = CRMMON_NODES_TAG_XML
        mock_set_crmmon_xml.return_value = None
        mock_get_node_state_tag_list.return_value = []
        mock_has_quorum.return_value = True

        obj = handle_host.HandleHost()

        self.assertRaisesRegex(
            Exception, "Failed to get nodes tag from crm_mon xml.",
            obj._check_host_status_by_crm_mon)
        mock_get_crmmon_xml.assert_called_once_with()
        mock_set_crmmon_xml.assert_called_once_with(CRMMON_NODES_TAG_XML)
        mock_get_node_state_tag_list.assert_called_once_with()

    @mock.patch.object(handle_host.HandleHost, '_get_crmmon_xml')
    def test_check_host_status_by_crm_mon_xml_is_None(
        self, mock_get_crmmon_xml):
        mock_get_crmmon_xml.return_value = None

        obj = handle_host.HandleHost()
        ret = obj._check_host_status_by_crm_mon()

        self.assertEqual(1, ret)
        mock_get_crmmon_xml.assert_called_once_with()

    @mock.patch.object(handle_host.HandleHost, '_check_if_status_changed')
    @mock.patch.object(parse_cib_xml.ParseCibXml, 'get_node_state_tag_list')
    @mock.patch.object(parse_cib_xml.ParseCibXml, 'have_quorum')
    @mock.patch.object(parse_cib_xml.ParseCibXml, 'set_cib_xml')
    @mock.patch.object(handle_host.HandleHost, '_get_cib_xml')
    def test_check_host_status_by_cibadmin(
        self, mock_get_cib_xml, mock_set_cib_xml, mock_have_quorum,
        mock_get_node_state_tag_list, mock_check_if_status_changed):
        mock_get_cib_xml.return_value = STATUS_TAG_XML
        mock_set_cib_xml.return_value = None
        mock_have_quorum.return_value = 1
        status_tag = ElementTree.fromstring(STATUS_TAG_XML)
        node_state_tag_list = list(status_tag)
        mock_get_node_state_tag_list.return_value = node_state_tag_list
        mock_check_if_status_changed.return_value = None

        obj = handle_host.HandleHost()
        ret = obj._check_host_status_by_cibadmin()

        self.assertEqual(0, ret)
        mock_get_cib_xml.assert_called_once_with()
        mock_set_cib_xml.assert_called_once_with(STATUS_TAG_XML)
        mock_have_quorum.assert_called_once_with()
        mock_get_node_state_tag_list.assert_called_once_with()
        mock_check_if_status_changed.assert_called_once_with(
            node_state_tag_list)

    @mock.patch.object(handle_host.HandleHost, '_check_if_status_changed')
    @mock.patch.object(parse_cib_xml.ParseCibXml, 'have_quorum')
    @mock.patch.object(parse_cib_xml.ParseCibXml, 'set_cib_xml')
    @mock.patch.object(handle_host.HandleHost, '_get_cib_xml')
    def test_check_host_status_by_cibadmin_no_quorum(
        self, mock_get_cib_xml, mock_set_cib_xml, mock_have_quorum,
        mock_check_if_status_changed):
        mock_get_cib_xml.return_value = STATUS_TAG_XML
        mock_set_cib_xml.return_value = None
        mock_have_quorum.return_value = 0

        obj = handle_host.HandleHost()
        ret = obj._check_host_status_by_cibadmin()

        self.assertEqual(2, ret)
        mock_get_cib_xml.assert_called_once_with()
        mock_set_cib_xml.assert_called_once_with(STATUS_TAG_XML)
        mock_have_quorum.assert_called_once_with()
        mock_check_if_status_changed.assert_not_called()

    @mock.patch.object(handle_host.HandleHost, '_get_cib_xml')
    def test_check_host_status_by_cibadmin_cib_xml_is_None(
        self, mock_get_cib_xml):
        mock_get_cib_xml.return_value = None

        obj = handle_host.HandleHost()
        ret = obj._check_host_status_by_cibadmin()

        self.assertEqual(1, ret)
        mock_get_cib_xml.assert_called_once_with()

    @mock.patch.object(parse_cib_xml.ParseCibXml, 'get_node_state_tag_list')
    @mock.patch.object(parse_cib_xml.ParseCibXml, 'have_quorum')
    @mock.patch.object(parse_cib_xml.ParseCibXml, 'set_cib_xml')
    @mock.patch.object(handle_host.HandleHost, '_get_cib_xml')
    def test_check_host_status_by_cibadmin_not_have_node_state_tag(
        self, mock_get_cib_xml, mock_set_cib_xml, mock_have_quorum,
        mock_get_node_state_tag_list):
        mock_get_cib_xml.return_value = STATUS_TAG_XML
        mock_set_cib_xml.return_value = None
        mock_have_quorum.return_value = 1
        mock_get_node_state_tag_list.return_value = []

        obj = handle_host.HandleHost()

        self.assertRaisesRegex(
            Exception, "Failed to get node_state tag from cib xml.",
            obj._check_host_status_by_cibadmin)
        mock_get_cib_xml.assert_called_once_with()
        mock_set_cib_xml.assert_called_once_with(STATUS_TAG_XML)
        mock_have_quorum.assert_called_once_with()
        mock_get_node_state_tag_list.assert_called_once_with()

    def test_stop(self):

        obj = handle_host.HandleHost()
        obj.stop()

        self.assertFalse(obj.running)

    @mock.patch.object(eventlet.greenthread, 'sleep')
    @mock.patch.object(handle_host.HandleHost,
                       '_check_host_status_by_cibadmin')
    @mock.patch.object(handle_host.HandleHost,
                       '_check_host_status_by_crmadmin')
    @mock.patch.object(handle_host.HandleHost, '_check_pacemaker_services')
    @mock.patch.object(handle_host.HandleHost, '_check_hb_line')
    def test_monitor_hosts(self,
                           mock_check_hb_line,
                           mock_check_pacemaker_services,
                           mock_check_host_status_by_crmadmin,
                           mock_check_host_status_by_cibadmin,
                           mock_sleep):

        mock_check_hb_line.side_effect = \
            [0, 1, 2, 0, Exception("Test exception."), 0, KeyboardInterrupt()]
        mock_check_pacemaker_services.side_effect = [True, False, False, True]
        mock_check_host_status_by_crmadmin.side_effect = [0, 1]
        mock_check_host_status_by_cibadmin.side_effect = [0, 1, 0]
        mock_sleep.return_value = None

        obj = handle_host.HandleHost()
        self.assertRaises(KeyboardInterrupt, obj.monitor_hosts)

        self.assertEqual(7, mock_check_hb_line.call_count)
        self.assertEqual(4, mock_check_pacemaker_services.call_count)
        mock_check_pacemaker_services.assert_called_with('pacemaker_remote')
        self.assertEqual(2, mock_check_host_status_by_crmadmin.call_count)
        self.assertEqual(3, mock_check_host_status_by_cibadmin.call_count)

    @mock.patch.object(eventlet.greenthread, 'sleep')
    @mock.patch.object(handle_host.HandleHost,
                       '_check_host_status_by_crm_mon')
    @mock.patch.object(handle_host.HandleHost, '_check_pacemaker_services')
    @mock.patch.object(handle_host.HandleHost, '_check_hb_line')
    def test_monitor_hosts_remotes_only(self,
                                        mock_check_hb_line,
                                        mock_check_pacemaker_services,
                                        mock_check_host_status_by_crm_mon,
                                        mock_sleep):

        CONF.host.restrict_to_remotes = True
        mock_check_hb_line.side_effect = \
            [0, Exception("Test exception."), 0, KeyboardInterrupt()]
        mock_check_pacemaker_services.return_value = True
        mock_check_host_status_by_crm_mon.return_value = 0
        mock_sleep.return_value = None

        obj = handle_host.HandleHost()
        self.assertRaises(KeyboardInterrupt, obj.monitor_hosts)

        self.assertEqual(4, mock_check_hb_line.call_count)
        self.assertEqual(2, mock_check_pacemaker_services.call_count)
        mock_check_pacemaker_services.assert_called_with('pacemaker_remote')
        self.assertEqual(2, mock_check_host_status_by_crm_mon.call_count)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from oslo_config import cfg


monitor_process_opts = [
    cfg.IntOpt('check_interval',
               default=5,
               help='Interval in seconds for checking a process.'),
    cfg.IntOpt('restart_retries',
               default=3,
               help='Number of retries when the failure of restarting a'
                    ' process.'),
    cfg.IntOpt('restart_interval',
               default=5,
               help='Interval in seconds for restarting a process.'),
    cfg.IntOpt('api_retry_max',
               default=12,
               help='Number of retries for send a notification in'
                    ' processmonitor.'),
    cfg.IntOpt('api_retry_interval',
               default=10,
               help='Interval between re-sending a notification in'
                    ' processmonitor(in seconds).'),
    cfg.StrOpt('process_list_path',
               default='/etc/masakarimonitors/process_list.yaml',
               help='The file path of process list.'),
]


def register_opts(conf):
    conf.register_opts(monitor_process_opts, group='process')


def list_opts():
    return {
        'process': monitor_process_opts
    }



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import eventlet
from keystoneauth1 import loading as ks_loading
from openstack import connection
from openstack import exceptions
from oslo_log import log as oslo_logging

import masakarimonitors.conf

LOG = oslo_logging.getLogger(__name__)
CONF = masakarimonitors.conf.CONF


class SendNotification(object):

    def __init__(self):
        self._masakari_client = None

    @property
    def masakari_client(self):
        if not self._masakari_client:
            self._masakari_client = self._make_client()
        return self._masakari_client

    def _make_client(self):
        auth = ks_loading.load_auth_from_conf_options(CONF, 'api')
        session = ks_loading.load_session_from_conf_options(CONF, 'api',
                                                            auth=auth)
        conn = connection.Connection(session=session,
                                     interface=CONF.api.api_interface,
                                     region_name=CONF.api.region)

        return conn.instance_ha

    def send_notification(self, api_retry_max, api_retry_interval, event):
        """Send a notification.

        This method sends a notification to the masakari-api.

        :param api_retry_max: Number of retries when the notification
            processing is error.
        :param api_retry_interval: Trial interval of time of the notification
            processing is error.
        :param event: dictionary of event that included in notification.
        """

        LOG.info("Send a notification. %s", event)

        # Send a notification.
        retry_count = 0
        while True:
            try:
                response = self.masakari_client.create_notification(
                    type=event['notification']['type'],
                    hostname=event['notification']['hostname'],
                    generated_time=event['notification']['generated_time'],
                    payload=event['notification']['payload'])

                LOG.info("Response: %s", response)
                break

            except Exception as e:
                if isinstance(e, exceptions.HttpException):
                    if e.status_code in [400, 409]:
                        LOG.info(e)
                        break

                if retry_count < api_retry_max:
                    LOG.warning("Retry sending a notification. (%s)", e)
                    retry_count = retry_count + 1
                    eventlet.greenthread.sleep(api_retry_interval)
                else:
                    LOG.exception("Exception caught: %s", e)
                    self._masakari_client = None
                    break



# Copyright (c) 2018 WindRiver Systems
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The code related to integration between oslo.cache module and masakarimonitors.

Use 'oslo_cache.dict'

i.e. dogpile.cache backend that uses dictionary for storage
Dogpile consists of two subsystems, one building on top of the other.
dogpile provides the concept of a dogpile lock, a control structure
which allows a single thread of execution to be selected as the creator
of some resource, while allowing other threads of execution to refer
to the previous version of this resource as the creation proceeds;
if there is no previous version, then those threads block until
the object is available.
"""
from oslo_cache import core
from oslo_config import cfg

from masakarimonitors.i18n import _

WEEK = 604800


def register_cache_configurations(conf):
    """Register all configurations required for oslo.cache.

    The procedure registers all configurations required for oslo.cache.
    It should be called before configuring of cache region

    :param conf: instance of configuration
    :returns: updated configuration
    """
    # register global configurations for caching
    core.configure(conf)

    # register specific configurations
    constraint_cache_group = cfg.OptGroup('constraint_validation_cache')
    constraint_cache_opts = [
        cfg.IntOpt('expiration_time', default=WEEK,
                   help=_(
                       'TTL, in seconds, for any cached item in the '
                       'dogpile.cache region used for caching of validation '
                       'constraints.')),
        cfg.BoolOpt("caching", default=True,
                    help=_(
                        'Toggle to enable/disable caching when Orchestration '
                        'Engine validates property constraints of stack.'
                        'During property validation with constraints '
                        'Orchestration Engine caches requests to other '
                        'OpenStack services. Please note that the global '
                        'toggle for oslo.cache(enabled=True in [cache] group) '
                        'must be enabled to use this feature.'))
    ]
    conf.register_group(constraint_cache_group)
    conf.register_opts(constraint_cache_opts, group=constraint_cache_group)

    extension_cache_group = cfg.OptGroup('service_extension_cache')
    extension_cache_opts = [
        cfg.IntOpt('expiration_time', default=WEEK,
                   help=_(
                       'TTL, in seconds, for any cached item in the '
                       'dogpile.cache region used for caching of service '
                       'extensions.')),
        cfg.BoolOpt('caching', default=True,
                    help=_(
                        'Toggle to enable/disable caching when Orchestration '
                        'Engine retrieves extensions from other OpenStack '
                        'services. Please note that the global toggle for '
                        'oslo.cache(enabled=True in [cache] group) must be '
                        'enabled to use this feature.'))
    ]
    conf.register_group(extension_cache_group)
    conf.register_opts(extension_cache_opts, group=extension_cache_group)

    find_cache_group = cfg.OptGroup('resource_finder_cache')
    find_cache_opts = [
        cfg.IntOpt('expiration_time', default=WEEK,
                   help=_(
                       'TTL, in seconds, for any cached item in the '
                       'dogpile.cache region used for caching of OpenStack '
                       'service finder functions.')),
        cfg.BoolOpt('caching', default=True,
                    help=_(
                        'Toggle to enable/disable caching when Orchestration '
                        'Engine looks for other OpenStack service resources '
                        'using name or id. Please note that the global '
                        'toggle for oslo.cache(enabled=True in [cache] group) '
                        'must be enabled to use this feature.'))
    ]
    conf.register_group(find_cache_group)
    conf.register_opts(find_cache_opts, group=find_cache_group)

    return conf


# variable that stores an initialized cache region
_REGION = None


def get_cache_region():
    global _REGION
    if not _REGION:
        _REGION = core.create_region()
        _REGION.configure('oslo_cache.dict',
            arguments={'expiration_time': WEEK})
        core.configure_cache_region(
            conf=register_cache_configurations(cfg.CONF),
            region=_REGION)
    return _REGION



# Copyright(c) 2024 Samsung SDS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from oslo_config import cfg

kubernetes_opts = [
    cfg.StrOpt('monitoring_node_labels',
               default='',
               help='''
Options for identifying the nodes to monitor. Only nodes with node labels
defined in this option are included in the monitor.

Example value:

monitoring_node_labels = "monitoring=true,compute-node=enabled,..."
'''),
]


def register_opts(conf):
    conf.register_opts(kubernetes_opts, group='kubernetes')


def list_opts():
    return {
        'kubernetes': kubernetes_opts
    }



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
import sys
import threading

from oslo_log import log as oslo_logging
from oslo_utils import excutils
from oslo_utils import timeutils

import masakarimonitors.conf
from masakarimonitors.instancemonitor.libvirt_handler import callback
from masakarimonitors.instancemonitor.libvirt_handler \
    import eventfilter_table as evft
from masakarimonitors.objects import event_constants as ec


LOG = oslo_logging.getLogger(__name__)
CONF = masakarimonitors.conf.CONF


class EventFilter(object):
    """Class of filtering events."""

    def __init__(self):
        self.callback = callback.Callback()

    def vir_event_filter(self, eventID, eventType, detail, uuID):
        """Filter events from libvirt.

        :param eventID: EventID
        :param eventType: Event type
        :param detail: Event name
        :pram uuID: UUID
        """

        noticeType = ec.EventConstants.TYPE_VM
        hostname = CONF.hostname or socket.gethostname()
        currentTime = timeutils.utcnow()

        # All Event Output if debug mode is on.
        msg = "libvirt Event Received.type = %s \
            hostname = %s uuid = %s time = %s eventID = %d eventType = %d \
            detail = %d" % (
            noticeType,
            hostname, uuID, currentTime, eventID,
            eventType, detail)
        LOG.debug(msg)

        try:
            if detail in evft.event_filter_dic[eventID][eventType]:
                LOG.debug("Event Filter Matched.")

                eventID_val = evft.eventID_dic[eventID]
                detail_val = evft.detail_dic[eventID][eventType][detail]

                # callback Thread Start
                thread = threading.Thread(
                    target=self.callback.libvirt_event_callback,
                    args=(eventID_val, detail_val,
                          uuID, noticeType,
                          hostname, currentTime)
                )
                thread.start()
            else:
                LOG.debug("Event Filter Unmatched.")

        except KeyError:
            LOG.debug("virEventFilter KeyError")
        except IndexError:
            LOG.debug("virEventFilter IndexError")
        except TypeError:
            LOG.debug("virEventFilter TypeError")
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.debug("Unexpected error: %s" % sys.exc_info()[0])



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from oslo_config import cfg
from oslo_config import types

monitor_host_opts = [
    cfg.StrOpt('monitoring_driver',
               default='default',
               help='Driver that hostmonitor uses for monitoring hosts.'),
    cfg.IntOpt('monitoring_interval',
               default=60,
               help='Monitoring interval(in seconds) of node status.'),
    cfg.IntOpt('monitoring_samples',
               default=1,
               help='''
Monitoring probes to collect before making the decision to send Masakari
notification about the node status. If and only if ``monitoring_samples``
consecutive reports have the same status, will the Masakari notification
be sent.
'''),
    cfg.IntOpt('api_retry_max',
               default=12,
               help='Number of retries for send a notification in'
                    ' hostmonitor.'),
    cfg.IntOpt('api_retry_interval',
               default=10,
               help='Trial interval of time of the notification processing'
                    ' is error(in seconds).'),
    cfg.BoolOpt('disable_ipmi_check',
                default=False,
                help='''
Do not check whether the host is completely down.

Possible values:

* True: Do not check whether the host is completely down.
* False: Do check whether the host is completely down.

If ipmi RA is not set in pacemaker, this value should be set True.
'''),
    cfg.BoolOpt('restrict_to_remotes',
               default=False,
               help='Only monitor pacemaker-remotes, ignore the status of'
                    ' full cluster members.'),
    cfg.IntOpt('ipmi_timeout',
               default=5,
               help='Timeout value(in seconds) of the ipmitool command.'),
    cfg.IntOpt('ipmi_retry_max',
               default=3,
               help='Number of ipmitool command retries.'),
    cfg.IntOpt('ipmi_retry_interval',
               default=10,
               help='Retry interval(in seconds) of the ipmitool command.'),
    cfg.IntOpt('stonith_wait',
               default=30,
               help='Standby time(in seconds) until activate STONITH.'),
    cfg.IntOpt('tcpdump_timeout',
               default=5,
               help='Timeout value(in seconds) of the tcpdump command when'
                    ' monitors the corosync communication.'),
    cfg.ListOpt('corosync_multicast_interfaces',
                help='''
The name of interface that corosync is using for mutual communication
between hosts.
If there are multiple interfaces, specify them in comma-separated
like 'enp0s3,enp0s8'.
The number of interfaces you specify must be equal to the number of
corosync_multicast_ports values and must be in correct order with relevant
ports in corosync_multicast_ports.
'''),
    cfg.ListOpt('corosync_multicast_ports',
                item_type=types.Port(),
                help='''
The port numbers that corosync is using for mutual communication
between hosts.
If there are multiple port numbers, specify them in comma-separated
like '5405,5406'.
The number of port numbers you specify must be equal to the number of
corosync_multicast_interfaces values and must be in correct order with
relevant interfaces in corosync_multicast_interfaces.
'''),
    cfg.StrOpt('pacemaker_node_type',
               default='autodetect',
               choices=('autodetect', 'cluster', 'remote'),
               help='''
Using this option, one can avoid systemd checks that would establish whether
this hostmonitor is running alongside Corosync and Pacemaker (the ``cluster``
stack) or Pacemaker Remote (the ``remote`` stack).

The default (``autodetect``) ensures backward compatibility and means systemd
is used to check the stack.
'''),
]


def register_opts(conf):
    conf.register_opts(monitor_host_opts, group='host')


def list_opts():
    return {
        'host': monitor_host_opts
    }



# Copyright (c) 2018 WindRiver Systems
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import eventlet
import logging

from oslo_config import cfg
from oslo_service import threadgroup

from masakarimonitors.introspectiveinstancemonitor import qemu_utils


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class ThreadGroupManager(object):
    """Thread group manager."""

    def init_qemu_ga(self):
        self.qemuGA = qemu_utils.QemuGuestAgent()
        LOG.debug('Started QemuGuestAgent')

    def __init__(self):
        self.init_qemu_ga()
        super(ThreadGroupManager, self).__init__()
        self.threads = {}
        self.group = threadgroup.ThreadGroup()

        # Create dummy service task, because when there is nothing queued
        # on self.tg the process exits
        self.add_timer(
            CONF.introspectiveinstancemonitor.guest_monitoring_interval,
            self._service_task)

    def _service_task(self):
        # This is used to trigger periodic monitoring tasks

        self.qemuGA.checkGuests()

    def start(self, func, *args, **kwargs):
        """Run the given method in a sub-thread."""

        LOG.debug('add_thread')
        return self.group.add_thread(func, *args, **kwargs)

    def add_timer(self, interval, func, *args, **kwargs):
        """Define a periodic task to be run in the thread group.

        The task will be executed in a separate green thread.
        Interval is from cfg.CONF.periodic_interval
        """

        LOG.debug('group.add_timer')
        self.group.add_timer(interval, func, *args, **kwargs)

    def stop_timers(self):
        self.group.stop_timers()

    def stop(self, graceful=False):
        """Stop any active threads belong to this threadgroup."""

        # Try to stop all threads gracefully
        self.group.stop(graceful)
        self.group.wait()

        # Wait for link()ed functions (i.e. lock release)
        threads = self.group.threads[:]
        links_done = dict((th, False) for th in threads)

        def mark_done(gt, th):
            links_done[th] = True

        for th in threads:
            th.link(mark_done, th)

        while not all(links_done.values()):
            eventlet.sleep()


def reschedule(action, sleep_time=1):
    """Eventlet Sleep for the specified number of seconds.

    :param sleep_time: seconds to sleep; if None, no sleep;
    """

    LOG.debug('At reschedule')
    if sleep_time is not None:
        LOG.debug('Action %s sleep for %s seconds' % (
            action.id, sleep_time))
        eventlet.sleep(sleep_time)


def sleep(sleep_time):
    """Interface for sleeping."""

    LOG.debug('sleep')

    eventlet.sleep(sleep_time)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class HostHoldStatus(object):
    """Hold host status.

    This class holds the host status.
    """

    def __init__(self):
        # host_status is dictionary like {'hostname': 'online'}.
        self.host_status = {}

    def set_host_status(self, node_state_tag):
        """Setter method.

        This method set host status by node_state tag of cib xml.

        :params node_state_tag: node_state tag of cib xml.
        """
        self.host_status[node_state_tag.get('uname')] = \
            node_state_tag.get('crmd')

    def get_host_status(self, hostname):
        """Getter method.

        This method returns the requested host status.
        host status is 'online' or 'offline'.

        :params hostname: Hostname to get status.

        :returns: Host status.
        """
        return self.host_status.get(hostname)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools
from unittest import mock
import yaml

import eventlet

from masakarimonitors.processmonitor import process as processmonitor_manager
from masakarimonitors.processmonitor.process_handler import handle_process

eventlet.monkey_patch(os=False)

MOCK_PROCESS_LIST = [
    {
        'process_name': 'mock_process_name_A',
        'start_command': 'mock_start_command',
        'pre_start_command': 'mock_pre_start_command',
        'post_start_command': 'mock_post_start_command',
        'restart_command': 'mock_restart_command',
        'pre_restart_command': 'mock_pre_restart_command',
        'post_restart_command': 'mock_post_restart_command',
        'run_as_root': True
    },
    {
        'id': 2,
        'process_name': 'mock_process_name_B',
        'start_command': 'mock_start_command',
        'pre_start_command': 'mock_pre_start_command',
        'post_start_command': 'mock_post_start_command',
        'restart_command': 'mock_restart_command',
        'pre_restart_command': 'mock_pre_restart_command',
        'post_restart_command': 'mock_post_restart_command',
        'run_as_root': True
    },
]

MOCK_DOWN_PROCESS_LIST = [
    {
        'process_name': 'mock_process_name_A',
        'start_command': 'mock_start_command',
        'pre_start_command': 'mock_pre_start_command',
        'post_start_command': 'mock_post_start_command',
        'restart_command': 'mock_restart_command',
        'pre_restart_command': 'mock_pre_restart_command',
        'post_restart_command': 'mock_post_restart_command',
        'run_as_root': True
    },
]


class TestProcessmonitorManager(testtools.TestCase):

    def setUp(self):
        super(TestProcessmonitorManager, self).setUp()

    def _get_mock_process_list(self, call_count):
        if call_count == 0:
            return MOCK_PROCESS_LIST
        else:
            return

    @mock.patch.object(eventlet.greenthread, 'sleep')
    @mock.patch.object(handle_process.HandleProcess, 'restart_processes')
    @mock.patch.object(handle_process.HandleProcess, 'monitor_processes')
    @mock.patch.object(handle_process.HandleProcess, 'start_processes')
    @mock.patch.object(handle_process.HandleProcess, 'set_process_list')
    @mock.patch.object(yaml, 'safe_load')
    @mock.patch('builtins.open')
    def test_main(self,
                  mock_file,
                  mock_load,
                  mock_set_process_list,
                  mock_start_processes,
                  mock_monitor_processes,
                  mock_restart_processes,
                  mock_sleep):

        mock_load.side_effect = [self._get_mock_process_list(0),
                                 self._get_mock_process_list(0),
                                 self._get_mock_process_list(1)]
        mock_set_process_list.return_value = None
        mock_start_processes.return_value = None
        mock_monitor_processes.side_effect = [MOCK_DOWN_PROCESS_LIST, []]
        mock_restart_processes.return_value = None
        mock_sleep.return_value = None

        obj = processmonitor_manager.ProcessmonitorManager()
        obj.main()

        mock_set_process_list.assert_called_with(MOCK_PROCESS_LIST)
        mock_start_processes.assert_called_once_with()
        self.assertEqual(2, mock_monitor_processes.call_count)
        mock_restart_processes.assert_called_once_with(MOCK_DOWN_PROCESS_LIST)

    @mock.patch.object(handle_process.HandleProcess, 'restart_processes')
    @mock.patch.object(handle_process.HandleProcess, 'monitor_processes')
    @mock.patch.object(handle_process.HandleProcess, 'start_processes')
    @mock.patch.object(handle_process.HandleProcess, 'set_process_list')
    @mock.patch.object(yaml, 'safe_load')
    @mock.patch('builtins.open')
    def test_main_exception(self,
                            mock_file,
                            mock_load,
                            mock_set_process_list,
                            mock_start_processes,
                            mock_monitor_processes,
                            mock_restart_processes):

        mock_load.return_value = self._get_mock_process_list(0)
        mock_set_process_list.return_value = None
        mock_start_processes.side_effect = Exception("Test exception.")

        obj = processmonitor_manager.ProcessmonitorManager()
        obj.main()

        mock_set_process_list.assert_called_once_with(MOCK_PROCESS_LIST)
        mock_start_processes.assert_called_once_with()
        mock_monitor_processes.assert_not_called()
        mock_restart_processes.assert_not_called()

    @mock.patch.object(handle_process.HandleProcess, 'set_process_list')
    @mock.patch.object(yaml, 'safe_load')
    @mock.patch('builtins.open')
    def test_load_process_list_yaml_error(self,
                                         mock_file,
                                         mock_load,
                                         mock_set_process_list):

        mock_load.side_effect = yaml.YAMLError

        obj = processmonitor_manager.ProcessmonitorManager()
        obj.main()

        mock_set_process_list.assert_not_called()

    @mock.patch.object(handle_process.HandleProcess, 'set_process_list')
    @mock.patch.object(yaml, 'safe_load')
    @mock.patch('builtins.open')
    def test_load_process_list_exception(self,
                                         mock_file,
                                         mock_load,
                                         mock_set_process_list):

        mock_load.side_effect = Exception("Test exception.")

        obj = processmonitor_manager.ProcessmonitorManager()
        obj.main()

        mock_set_process_list.assert_not_called()

    def test_stop(self):

        obj = processmonitor_manager.ProcessmonitorManager()
        obj.stop()

        self.assertFalse(obj.running)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
import testtools
from unittest import mock
import uuid

import eventlet
from oslo_utils import timeutils

from masakarimonitors.ha import masakari
from masakarimonitors.instancemonitor.libvirt_handler import callback
from masakarimonitors.objects import event_constants as ec

eventlet.monkey_patch(os=False)


class TestCallback(testtools.TestCase):

    def setUp(self):
        super(TestCallback, self).setUp()

    @mock.patch.object(masakari.SendNotification, 'send_notification')
    def test_libvirt_event_callback(self, mock_send_notification):

        mock_send_notification.return_value = None

        obj = callback.Callback()

        event_id = 0
        details = 5
        domain_uuid = uuid.uuid4()
        notice_type = ec.EventConstants.TYPE_VM
        hostname = socket.gethostname()
        current_time = timeutils.utcnow()

        obj.libvirt_event_callback(event_id, details, domain_uuid,
            notice_type, hostname, current_time)

        retry_max = 12
        retry_interval = 10
        event = {
            'notification': {
                'type': notice_type,
                'hostname': hostname,
                'generated_time': current_time,
                'payload': {
                    'event': event_id,
                    'instance_uuid': domain_uuid,
                    'vir_domain_event': details
                }
            }
        }
        mock_send_notification.assert_called_once_with(
            retry_max, retry_interval, event)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc


class DriverBase(object, metaclass=abc.ABCMeta):
    """Driver Base class.

    This class is base of monitoring hosts.
    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def monitor_hosts(self):
        """Must override monitor_hosts method."""
        pass



# Copyright(c) 2021 Inspur
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile

from oslo_config import fixture as fixture_config
import testtools
import yaml

from masakarimonitors.hostmonitor.consul_check import matrix_helper


class TestMatrixManager(testtools.TestCase):

    def setUp(self):
        super(TestMatrixManager, self).setUp()
        self.CONF = self.useFixture(fixture_config.Config()).conf

    def test_get_matrix_and_sequence_from_file(self):
        matrix_cfg = {
            'sequence': ['manage', 'tenant', 'storage'],
            'matrix': [{"health": ["up", "up", "up"],
                        "action": ['test']}]
        }
        tmp_cfg = tempfile.NamedTemporaryFile(mode='w', delete=False)
        tmp_cfg.write(yaml.safe_dump(matrix_cfg))
        tmp_cfg.close()
        self.CONF.set_override('matrix_config_file',
                               tmp_cfg.name, group='consul')

        matrix_manager = matrix_helper.MatrixManager(self.CONF)
        self.assertEqual(matrix_cfg.get('sequence'),
                         matrix_manager.get_sequence())
        self.assertEqual(matrix_cfg.get('matrix'),
                         matrix_manager.get_matrix())

    def test_get_default_matrix_and_sequence(self):
        self.CONF.set_override('matrix_config_file', None, group='consul')

        matrix_manager = matrix_helper.MatrixManager(self.CONF)
        self.assertEqual(matrix_helper.DEFAULT_SEQUENCE,
                         matrix_manager.get_sequence())
        self.assertEqual(matrix_helper.DEFAULT_MATRIX,
                         matrix_manager.get_matrix())



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket

from collections import deque
import eventlet
from oslo_log import log as oslo_logging
from oslo_utils import timeutils

import masakarimonitors.conf
from masakarimonitors.ha import masakari
import masakarimonitors.hostmonitor.driver as driver
from masakarimonitors.hostmonitor.host_handler import hold_host_status
from masakarimonitors.hostmonitor.host_handler import parse_cib_xml
from masakarimonitors.hostmonitor.host_handler import parse_crmmon_xml
from masakarimonitors.objects import event_constants as ec
from masakarimonitors import utils

LOG = oslo_logging.getLogger(__name__)
CONF = masakarimonitors.conf.CONF


class CibSchemaCompliantTag(dict):
    """Create a dict which has the same attributes as a cib node tag.

    Given a crm node tag convert it to a dict with corresponding cib tag
    attributes.
    """
    def __init__(self, crmon_entry):
        self['uname'] = crmon_entry.get('name')
        online = crmon_entry.get('online')
        self['crmd'] = 'online' if online == 'true' else 'offline'


class HandleHost(driver.DriverBase):
    """Handle hosts.

    This class handles the host status.
    """

    def __init__(self):
        super(HandleHost, self).__init__()
        self.my_hostname = socket.gethostname()
        self.xml_parser = parse_cib_xml.ParseCibXml()
        self.crmmon_xml_parser = parse_crmmon_xml.ParseCrmMonXml()
        self.status_holder = hold_host_status.HostHoldStatus()
        self.notifier = masakari.SendNotification()
        self.monitoring_data = {}

    def _update_monitoring_data(self, hostname, status):
        health_history = self.monitoring_data.setdefault(
            hostname, deque([], maxlen=CONF.host.monitoring_samples))
        health_history.append(status)

    def get_stabilised_host_status(self, hostname):
        health_history = self.monitoring_data.get(hostname)
        if len(health_history) < CONF.host.monitoring_samples:
            LOG.debug("Not enough monitoring data for host %s.", hostname)
            return '_being_collected'

        stabilised_status = health_history[0]

        # If and only if the sequence of host status is consistently the same,
        # will it return that status.
        if len(health_history) == health_history.count(stabilised_status):
            return stabilised_status
        else:
            return '_uncertain'

    def _check_pacemaker_services(self, target_service):
        try:
            cmd_str = 'systemctl status ' + target_service
            command = cmd_str.split()

            # Execute command.
            out, err = utils.execute(*command, run_as_root=True)

            if err:
                raise Exception("Service check for %s returned errors: %s"
                % (target_service, err))

            return True

        except Exception:
            LOG.warning("Failed to check pacemaker service '%s'.",
                        target_service, exc_info=True)
            return False

    def _check_hb_line(self):
        """Check whether the corosync communication is normal.

        :returns: 0 if normal, 1 if abnormal, 2 if configuration file is
            wrong or neither pacemaker nor pacemaker-remote is running.
        """
        if CONF.host.pacemaker_node_type == 'autodetect':
            # Check whether the Corosync/Pacemaker services are running.
            corosync_status = self._check_pacemaker_services('corosync')
            pacemaker_status = self._check_pacemaker_services('pacemaker')
            pacemaker_remote_status = self._check_pacemaker_services(
                'pacemaker_remote')
        else:
            corosync_status = (
                CONF.host.pacemaker_node_type == 'cluster')
            pacemaker_status = (
                CONF.host.pacemaker_node_type == 'cluster')
            pacemaker_remote_status = (
                CONF.host.pacemaker_node_type == 'remote')

        if corosync_status is False or pacemaker_status is False:
            if pacemaker_remote_status is False:
                LOG.error(
                    "Neither pacemaker nor pacemaker-remote is running.")
                return 2
            else:
                LOG.info("Works on pacemaker-remote.")
                return 0

        # Check whether the necessary parameters are set.
        if not CONF.host.corosync_multicast_interfaces or \
                not CONF.host.corosync_multicast_ports:
            msg = ("corosync_multicast_interfaces or "
                   "corosync_multicast_ports is not set.")
            LOG.error("%s", msg)
            return 2

        # Check whether the corosync communication is normal.
        corosync_multicast_interfaces = CONF.host.corosync_multicast_interfaces
        corosync_multicast_ports = CONF.host.corosync_multicast_ports

        if len(corosync_multicast_interfaces) != len(corosync_multicast_ports):
            msg = ("Incorrect parameters corosync_multicast_interfaces or "
                   "corosync_multicast_ports.")
            LOG.error("%s", msg)
            return 2

        is_nic_normal = False
        for num in range(0, len(corosync_multicast_interfaces)):
            cmd_str = ("timeout %s tcpdump -n -c 1 -p -i %s port %d") \
                % (CONF.host.tcpdump_timeout,
                   corosync_multicast_interfaces[num],
                   corosync_multicast_ports[num])
            command = cmd_str.split()

            try:
                # Execute tcpdump command.
                out, err = utils.execute(*command, run_as_root=True)

                # If command doesn't raise exception, nic is normal.
                msg = ("Corosync communication using '%s' is normal.") \
                    % corosync_multicast_interfaces[num]
                LOG.info("%s", msg)
                is_nic_normal = True
                break
            except Exception:
                msg = ("Corosync communication using '%s' is failed.") \
                    % corosync_multicast_interfaces[num]
                LOG.warning("%s", msg)

        if is_nic_normal is False:
            LOG.error("Corosync communication is failed.")
            return 1

        return 0

    def _check_host_status_by_crmadmin(self):
        try:
            # Execute crmadmin command.
            out, err = utils.execute('crmadmin', '-S', self.my_hostname,
                                     run_as_root=True)

            if err:
                msg = ("crmadmin command output stderr: %s") % err
                raise Exception(msg)

            # If own host is stable status, crmadmin outputs
            # 'S_IDLE' or 'S_NOT_DC'
            if 'S_IDLE' in out or 'S_NOT_DC' in out:
                return 0
            else:
                raise Exception(
                    "crmadmin command output unexpected host status.")

        except Exception as e:
            LOG.warning("Exception caught: %s", e)
            LOG.warning("'%s' is unstable state on cluster.",
                        self.my_hostname)
            return 1

    def _get_cib_xml(self):
        try:
            # Execute cibadmin command.
            out, err = utils.execute('cibadmin', '--query', run_as_root=True)

            if err:
                msg = ("cibadmin command output stderr: %s") % err
                raise Exception(msg)

        except Exception as e:
            LOG.warning("Exception caught: %s", e)
            return

        return out

    def _get_crmmon_xml(self):
        """Get summary of cluster's current state in XML format."""
        try:
            # Execute crm_mon command.
            out, err = utils.execute('crm_mon', '-X', run_as_root=True)

            if err:
                msg = ("crmmon command output stderr: %s") % err
                raise Exception(msg)

        except Exception as e:
            LOG.warning("Exception caught: %s", e)
            return

        return out

    def _is_poweroff(self, hostname):
        ipmi_values = self.xml_parser.get_stonith_ipmi_params(hostname)
        if ipmi_values is None:
            LOG.error("Failed to get params of ipmi RA.")
            return False

        cmd_str = ("timeout %s ipmitool -U %s -P %s -I %s -H %s "
                   "power status") \
            % (str(CONF.host.ipmi_timeout), ipmi_values['userid'],
               ipmi_values['passwd'], ipmi_values['interface'],
               ipmi_values['ipaddr'])
        command = cmd_str.split()

        retry_count = 0
        while True:
            try:
                # Execute ipmitool command.
                out, err = utils.execute(*command, run_as_root=False)

                if err:
                    msg = ("ipmitool command output stderr: %s") % err
                    raise Exception(msg)

                msg = ("ipmitool command output stdout: %s") % out

                if 'Power is off' in out:
                    LOG.info("%s", msg)
                    return True
                else:
                    raise Exception(msg)

            except Exception as e:
                if retry_count < CONF.host.ipmi_retry_max:
                    LOG.warning("Retry executing ipmitool command. (%s)", e)
                    retry_count = retry_count + 1
                    eventlet.greenthread.sleep(CONF.host.ipmi_retry_interval)
                else:
                    LOG.error("Exception caught: %s", e)
                    return False

    def _make_event(self, hostname, current_status):

        if current_status == 'online':
            # Set values that host has started.
            event_type = ec.EventConstants.EVENT_STARTED
            cluster_status = current_status.upper()
            host_status = ec.EventConstants.HOST_STATUS_NORMAL

        else:
            # Set values that host has stopped.
            event_type = ec.EventConstants.EVENT_STOPPED
            cluster_status = current_status.upper()

            if not CONF.host.disable_ipmi_check:
                if self._is_poweroff(hostname):
                    # Set value that host status is normal.
                    host_status = ec.EventConstants.HOST_STATUS_NORMAL
                else:
                    # Set value that host status is unknown.
                    host_status = ec.EventConstants.HOST_STATUS_UNKNOWN
            else:
                # Set value that host status is normal.
                host_status = ec.EventConstants.HOST_STATUS_NORMAL

        current_time = timeutils.utcnow()
        event = {
            'notification': {
                'type': ec.EventConstants.TYPE_COMPUTE_HOST,
                'hostname': hostname,
                'generated_time': current_time,
                'payload': {
                    'event': event_type,
                    'cluster_status': cluster_status,
                    'host_status': host_status
                }
            }
        }

        return event

    def _normalize_host_status(self, host_status):
        if host_status is None:
            return None

        try:
            return 'online' if int(host_status) > 0 else 'offline'
        except ValueError:
            return host_status

    def _check_if_status_changed(self, node_state_tag_list):

        # Check if host status changed.
        for node_state_tag in node_state_tag_list:
            hostname = node_state_tag.get('uname')
            # hostmonitor doesn't monitor itself.
            if hostname == self.my_hostname:
                continue

            current_status = self._normalize_host_status(
                node_state_tag.get('crmd')
            )
            self._update_monitoring_data(hostname, current_status)
            old_status = self._normalize_host_status(
                self.status_holder.get_host_status(hostname)
            )

            # If old_status is None, This is first get of host status.
            if old_status is None:
                msg = ("Recognized '%s' as a new member of cluster."
                       " Host status is '%s'.") \
                    % (hostname, current_status)
                LOG.info("%s", msg)
                self.status_holder.set_host_status(node_state_tag)
                continue

            stabilised_status = self.get_stabilised_host_status(hostname)

            # Output host status.
            msg = ("'%s' is '%s' (current: '%s').") % (hostname,
                                                       stabilised_status,
                                                       current_status)
            LOG.info("%s", msg)

            if stabilised_status == '_being_collected':
                continue

            # If host stabilised status changed, send a notification.
            if stabilised_status != old_status:
                if stabilised_status not in ['online', 'offline']:
                    # If stabilised_status is not 'online' or 'offline',
                    # hostmonitor doesn't send a notification.
                    msg = ("Since host status is '%s',"
                           " hostmonitor doesn't send a notification.") \
                        % stabilised_status
                    LOG.info("%s", msg)
                else:
                    event = self._make_event(hostname, stabilised_status)

                    # Send a notification.
                    self.notifier.send_notification(
                        CONF.host.api_retry_max,
                        CONF.host.api_retry_interval,
                        event)

                if stabilised_status != '_uncertain':
                    # Update host status.
                    self.status_holder.set_host_status(node_state_tag)

    def _check_host_status_by_crm_mon(self):
        crmmon_xml = self._get_crmmon_xml()
        if crmmon_xml is None:
            # crm_mon command failure.
            return 1

        # Set to the ParseCrmMonXml object.
        self.crmmon_xml_parser.set_crmmon_xml(crmmon_xml)

        # Check if the cluster has quorum.
        if not self.crmmon_xml_parser.has_quorum():
            msg = "Pacemaker cluster doesn't have quorum."
            LOG.warning("%s", msg)
            return 2

        # Get node_state tag list.
        node_state_tag_list = self.crmmon_xml_parser.get_node_state_tag_list()
        if len(node_state_tag_list) == 0:
            # If crmmon xml doesn't have node_state tag,
            # it is an unexpected result.
            raise Exception(
                "Failed to get nodes tag from crm_mon xml.")

        node_state_tag_list = [CibSchemaCompliantTag(n)
                               for n in node_state_tag_list
                               if n.get('type') == 'remote']
        # Check if status changed.
        self._check_if_status_changed(node_state_tag_list)

        return 0

    def _check_host_status_by_cibadmin(self):
        # Get xml of cib info.
        cib_xml = self._get_cib_xml()
        if cib_xml is None:
            # cibadmin command failure.
            return 1

        # Set to the ParseCibXml object.
        self.xml_parser.set_cib_xml(cib_xml)

        # Check if pacemaker cluster have quorum.
        if self.xml_parser.have_quorum() == 0:
            msg = "Pacemaker cluster doesn't have quorum."
            LOG.warning("%s", msg)
            return 2

        # Get node_state tag list.
        node_state_tag_list = self.xml_parser.get_node_state_tag_list()
        if len(node_state_tag_list) == 0:
            # If cib xml doesn't have node_state tag,
            # it is an unexpected result.
            raise Exception(
                "Failed to get node_state tag from cib xml.")

        # Check if status changed.
        self._check_if_status_changed(node_state_tag_list)

        return 0

    def stop(self):
        self.running = False

    def monitor_hosts(self):
        """Host monitoring main method.

        This method monitors hosts.
        """
        self.running = True
        while self.running:
            try:
                # Check whether corosync communication between hosts
                # is normal.
                ret = self._check_hb_line()
                if ret == 1:
                    # Because my host may be fenced by stonith due to split
                    # brain condition, sleep for a certain time.
                    eventlet.greenthread.sleep(CONF.host.stonith_wait)
                elif ret == 2:
                    LOG.warning("hostmonitor skips monitoring hosts.")
                    eventlet.greenthread.sleep(CONF.host.monitoring_interval)
                    continue

                if CONF.host.pacemaker_node_type == 'autodetect':
                    pacemaker_remote_status = self._check_pacemaker_services(
                        'pacemaker_remote')
                else:
                    pacemaker_remote_status = (
                        CONF.host.pacemaker_node_type == 'remote')

                # Check the host status is stable or unstable by crmadmin.
                # It only checks when this process runs on the full cluster
                # stack of corosync.
                if pacemaker_remote_status is False:
                    if self._check_host_status_by_crmadmin() != 0:
                        LOG.warning("hostmonitor skips monitoring hosts.")
                        eventlet.greenthread.sleep(
                            CONF.host.monitoring_interval)
                        continue

                # Check the host status is online or offline.
                if CONF.host.restrict_to_remotes:
                    status_func = self._check_host_status_by_crm_mon
                else:
                    status_func = self._check_host_status_by_cibadmin

                if status_func() != 0:
                    LOG.warning("hostmonitor skips monitoring hosts.")

            except Exception as e:
                LOG.exception("Exception caught: %s", e)

            eventlet.greenthread.sleep(CONF.host.monitoring_interval)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from oslo_log import log

import masakarimonitors.conf
from masakarimonitors import version


CONF = masakarimonitors.conf.CONF


def parse_args(argv, default_config_files=None):
    log.register_options(CONF)
    # We use the oslo.log default log levels which includes suds=INFO
    # and add only the extra levels that Masakari needs
    log.set_defaults(default_log_levels=log.get_default_log_levels())

    CONF(argv[1:],
         project='masakarimonitors',
         version=version.version_string(),
         default_config_files=default_config_files)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from oslo_config import cfg

from masakarimonitors.conf import api
from masakarimonitors.conf import base
from masakarimonitors.conf import consul
from masakarimonitors.conf import host
from masakarimonitors.conf import instance
from masakarimonitors.conf import introspectiveinstancemonitor
from masakarimonitors.conf import kubernetes
from masakarimonitors.conf import process
from masakarimonitors.conf import service

CONF = cfg.CONF

api.register_opts(CONF)
base.register_opts(CONF)
consul.register_opts(CONF)
host.register_opts(CONF)
instance.register_opts(CONF)
introspectiveinstancemonitor.register_opts(CONF)
kubernetes.register_opts(CONF)
process.register_opts(CONF)
service.register_opts(CONF)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from oslo_config import cfg


monitor_callback_opts = [
    cfg.IntOpt('retry_max',
               default=12,
               help='Number of retries when the notification processing'
                    ' is error.'),
    cfg.IntOpt('retry_interval',
               default=10,
               help='Trial interval of time of the notification processing'
                    ' is error(in seconds).'),
]

libvirt_opts = [
    cfg.StrOpt('connection_uri',
               default='qemu:///system',
               help='Override the default libvirt URI.')
]


def register_opts(conf):
    conf.register_opts(monitor_callback_opts, group='callback')
    conf.register_opts(libvirt_opts, group='libvirt')


def list_opts():
    return {
        'callback': monitor_callback_opts,
        'libvirt': libvirt_opts
    }



# Copyright(c) 2021 Inspur
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools
from unittest import mock

from oslo_config import fixture as fixture_config

from masakarimonitors.hostmonitor.consul_check import consul_helper


class FakerAgentMembers(object):

    def __init__(self):
        self.agent_members = []

    def create_agent(self, name, status=1):
        agent = {
            'Name': name,
            'Status': status,
            'Port': 'agent_lan_port',
            'Addr': 'agent_ip',
            'Tags': {
                'dc': 'storage',
                'role': 'consul',
                'port': 'agent_server_port',
                'wan_join_port': 'agent_wan_port',
                'expect': '3',
                'id': 'agent_id',
                'vsn_max': '3',
                'vsn_min': '2',
                'vsn': '2',
                'raft_vsn': '2',
            },
            'ProtocolMax': 5,
            'ProtocolMin': 1,
            'ProtocolCur': 2,
            'DelegateMax': 5,
            'DelegateMin': 2,
            'DelegateCur': 4,
        }

        self.agent_members.append(agent)

    def generate_agent_members(self):
        return self.agent_members


class TestConsulManager(testtools.TestCase):

    def setUp(self):
        super(TestConsulManager, self).setUp()
        self.CONF = self.useFixture(fixture_config.Config()).conf
        self.consul_manager = consul_helper.ConsulManager(self.CONF)
        self.consul_manager.agents = {
            'manage': consul_helper.ConsulAgent('manage'),
            'tenant': consul_helper.ConsulAgent('tenant'),
            'storage': consul_helper.ConsulAgent('storage'),
        }

    @mock.patch('consul.Consul')
    def test_init_agents_no_agent(self, mock_consul):
        mgr = consul_helper.ConsulManager(self.CONF)
        self.assertEqual({}, mgr.agents)
        mock_consul.assert_not_called()

    @mock.patch('consul.Consul')
    def test_init_agents_ipv4(self, mock_consul):
        self.CONF.set_override('agent_manage', '192.0.2.10:1000', 'consul')
        self.CONF.set_override('agent_tenant', '192.0.2.11:1001', 'consul')
        self.CONF.set_override('agent_storage', '192.0.2.12', 'consul')
        mgr = consul_helper.ConsulManager(self.CONF)
        self.assertEqual('manage', mgr.agents['manage'].name)
        self.assertEqual('192.0.2.10', mgr.agents['manage'].addr)
        self.assertEqual(1000, mgr.agents['manage'].port)
        self.assertEqual(mock.call(host='192.0.2.10', port=1000),
                         mock_consul.mock_calls[0])
        self.assertEqual('tenant', mgr.agents['tenant'].name)
        self.assertEqual('192.0.2.11', mgr.agents['tenant'].addr)
        self.assertEqual(1001, mgr.agents['tenant'].port)
        self.assertEqual(mock.call(host='192.0.2.11', port=1001),
                         mock_consul.mock_calls[1])
        self.assertEqual('storage', mgr.agents['storage'].name)
        self.assertEqual('192.0.2.12', mgr.agents['storage'].addr)
        self.assertEqual(8500, mgr.agents['storage'].port)
        self.assertEqual(mock.call(host='192.0.2.12', port=8500),
                         mock_consul.mock_calls[2])
        self.assertEqual(3, mock_consul.call_count)

    @mock.patch('consul.Consul')
    def test_init_agents_ipv6(self, mock_consul):
        self.CONF.set_override(
            'agent_manage', '[2001:db8::10]:1000', 'consul')
        self.CONF.set_override(
            'agent_tenant', '[2001:db8::11]:1001', 'consul')
        self.CONF.set_override(
            'agent_storage', '2001:db8::12', 'consul')
        mgr = consul_helper.ConsulManager(self.CONF)
        self.assertEqual('manage', mgr.agents['manage'].name)
        self.assertEqual('2001:db8::10', mgr.agents['manage'].addr)
        self.assertEqual(1000, mgr.agents['manage'].port)
        self.assertEqual(mock.call(host='2001:db8::10', port=1000),
                         mock_consul.mock_calls[0])
        self.assertEqual('tenant', mgr.agents['tenant'].name)
        self.assertEqual('2001:db8::11', mgr.agents['tenant'].addr)
        self.assertEqual(1001, mgr.agents['tenant'].port)
        self.assertEqual(mock.call(host='2001:db8::11', port=1001),
                         mock_consul.mock_calls[1])
        self.assertEqual('storage', mgr.agents['storage'].name)
        self.assertEqual('2001:db8::12', mgr.agents['storage'].addr)
        self.assertEqual(8500, mgr.agents['storage'].port)
        self.assertEqual(mock.call(host='2001:db8::12', port=8500),
                         mock_consul.mock_calls[2])
        self.assertEqual(3, mock_consul.call_count)

    def test_get_health(self):
        fake_manage_agents = FakerAgentMembers()
        fake_manage_agents.create_agent('node01', status=1)
        fake_manage_agents.create_agent('node02', status=1)
        fake_manage_agents.create_agent('node03', status=1)
        agent_manage_members = fake_manage_agents.generate_agent_members()

        fake_tenant_agents = FakerAgentMembers()
        fake_tenant_agents.create_agent('node01', status=1)
        fake_tenant_agents.create_agent('node02', status=1)
        fake_tenant_agents.create_agent('node03', status=1)
        agent_tenant_members = fake_tenant_agents.generate_agent_members()

        fake_storage_agents = FakerAgentMembers()
        fake_storage_agents.create_agent('node01', status=1)
        fake_storage_agents.create_agent('node02', status=1)
        fake_storage_agents.create_agent('node03', status=3)
        agent_storage_members = fake_storage_agents.generate_agent_members()

        with mock.patch.object(self.consul_manager.agents['manage'],
                'get_agents', return_value=agent_manage_members):
            with mock.patch.object(self.consul_manager.agents['tenant'],
                    'get_agents', return_value=agent_tenant_members):
                with mock.patch.object(self.consul_manager.agents['storage'],
                        'get_agents', return_value=agent_storage_members):
                    excepted_health = {
                        "node01": ['up', 'up', 'up'],
                        "node02": ['up', 'up', 'up'],
                        "node03": ['up', 'up', 'down'],
                    }
                    sequence = ['manage', 'tenant', 'storage']
                    agents_health = self.consul_manager.get_health(sequence)
                    self.assertEqual(excepted_health, agents_health)


class TestConsulAgent(testtools.TestCase):

    def setUp(self):
        super(TestConsulAgent, self).setUp()
        self.consul_agent = consul_helper.ConsulAgent('test')

    def test_get_health(self):
        fake_agents = FakerAgentMembers()
        fake_agents.create_agent('node01', status=1)
        fake_agents.create_agent('node02', status=1)
        fake_agents.create_agent('node03', status=3)
        agent_members = fake_agents.generate_agent_members()

        with mock.patch.object(self.consul_agent, 'get_agents',
                return_value=agent_members):
            excepted_health = {
                "node01": 'up',
                "node02": 'up',
                "node03": 'down',
            }
            agents_health = self.consul_agent.get_health()
            self.assertEqual(excepted_health, agents_health)



# Copyright(c) 2017 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import testtools
from unittest import mock

from stevedore import driver

from masakarimonitors.hostmonitor import host


class TestHostmonitorManager(testtools.TestCase):

    def setUp(self):
        super(TestHostmonitorManager, self).setUp()

    @mock.patch.object(driver, 'DriverManager')
    def test_init_host(self, mock_DriverManager):

        mock_driver = mock.Mock()
        mock_DriverManager.return_value = mock_driver

        host_manager = host.HostmonitorManager()
        host_manager.init_host()

        mock_DriverManager.assert_called_once_with(
            namespace='hostmonitor.driver',
            name='default',
            invoke_on_load=True,
            invoke_args=(),
        )

    @mock.patch.object(os, '_exit')
    @mock.patch.object(driver, 'DriverManager')
    def test_init_host_exception(self, mock_DriverManager, mock_exit):

        mock_DriverManager.side_effect = Exception("Test exception.")
        mock_exit.return_value = None

        host_manager = host.HostmonitorManager()
        host_manager.init_host()

        mock_DriverManager.assert_called_once_with(
            namespace='hostmonitor.driver',
            name='default',
            invoke_on_load=True,
            invoke_args=(),
        )
        mock_exit.assert_called_once_with(1)

    @mock.patch.object(driver, 'DriverManager')
    def test_stop(self, mock_DriverManager):

        mock_driver = mock.Mock()
        mock_DriverManager.return_value = mock_driver

        host_manager = host.HostmonitorManager()
        host_manager.init_host()
        host_manager.stop()

        mock_DriverManager.assert_called_once_with(
            namespace='hostmonitor.driver',
            name='default',
            invoke_on_load=True,
            invoke_args=(),
        )
        mock_driver.driver.stop.assert_called_once_with()

    @mock.patch.object(driver, 'DriverManager')
    def test_main(self, mock_DriverManager):

        mock_driver = mock.Mock()
        mock_DriverManager.return_value = mock_driver

        host_manager = host.HostmonitorManager()
        host_manager.init_host()
        ret = host_manager.main()

        mock_DriverManager.assert_called_once_with(
            namespace='hostmonitor.driver',
            name='default',
            invoke_on_load=True,
            invoke_args=(),
        )
        mock_driver.driver.monitor_hosts.assert_called_once_with()
        self.assertIsNone(ret)

    @mock.patch.object(driver, 'DriverManager')
    def test_main_exception(self, mock_DriverManager):

        mock_driver = mock.Mock()
        mock_DriverManager.return_value = mock_driver
        mock_driver.driver.monitor_hosts.side_effect = \
            Exception("Test exception.")

        host_manager = host.HostmonitorManager()
        host_manager.init_host()
        ret = host_manager.main()

        mock_DriverManager.assert_called_once_with(
            namespace='hostmonitor.driver',
            name='default',
            invoke_on_load=True,
            invoke_args=(),
        )
        mock_driver.driver.monitor_hosts.assert_called_once_with()
        self.assertIsNone(ret)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from stevedore import driver

from oslo_log import log as oslo_logging

import masakarimonitors.conf
from masakarimonitors import manager

LOG = oslo_logging.getLogger(__name__)
CONF = masakarimonitors.conf.CONF


class HostmonitorManager(manager.Manager):
    """Manages the masakari-hostmonitor."""

    def __init__(self, *args, **kwargs):
        super(HostmonitorManager, self).__init__(
            service_name="hostmonitor", *args, **kwargs)
        self.driver = None

    def init_host(self):
        """Initialization for hostmonitor."""
        try:
            # Determine dynamic load driver from configuration.
            driver_name = CONF.host.monitoring_driver

            # Load the driver to global.
            self.driver = driver.DriverManager(
                namespace='hostmonitor.driver',
                name=driver_name,
                invoke_on_load=True,
                invoke_args=(),
            )
        except Exception as e:
            LOG.exception(
                "Exception caught during initializing hostmonitor: %s", e)
            os._exit(1)

    def stop(self):
        self.driver.driver.stop()

    def main(self):
        """Main method."""

        try:
            # Call the host monitoring driver.
            self.driver.driver.monitor_hosts()

        except Exception as e:
            LOG.exception("Exception caught: %s", e)

        return



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools
from xml.etree import ElementTree

import eventlet

from masakarimonitors.hostmonitor.host_handler import hold_host_status

eventlet.monkey_patch(os=False)

NODE_STATE_XML = '<node_state uname="masakari-node" crmd="online">' \
                 '  <test foo="foo"/>' \
                 '</node_state>'
NODE_STATE_TAG = ElementTree.fromstring(NODE_STATE_XML)


class TestHostHoldStatus(testtools.TestCase):

    def setUp(self):
        super(TestHostHoldStatus, self).setUp()

    def test_set_host_status(self):
        obj = hold_host_status.HostHoldStatus()
        obj.set_host_status(NODE_STATE_TAG)
        self.assertEqual('online', obj.get_host_status('masakari-node'))

    def test_get_host_status(self):
        obj = hold_host_status.HostHoldStatus()
        obj.set_host_status(NODE_STATE_TAG)
        self.assertEqual('online', obj.get_host_status('masakari-node'))
